"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
If you make changes to this file, run "python -m dagster._grpc.compile" after."""

import builtins
import google.protobuf.descriptor
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class Empty(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___Empty = Empty

@typing_extensions.final
class PingRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ECHO_FIELD_NUMBER: builtins.int
    echo: builtins.str
    def __init__(
        self,
        *,
        echo: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["echo", b"echo"]) -> None: ...

global___PingRequest = PingRequest

@typing_extensions.final
class PingReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ECHO_FIELD_NUMBER: builtins.int
    SERIALIZED_SERVER_UTILIZATION_METRICS_FIELD_NUMBER: builtins.int
    echo: builtins.str
    serialized_server_utilization_metrics: builtins.str
    def __init__(
        self,
        *,
        echo: builtins.str = ...,
        serialized_server_utilization_metrics: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "echo",
            b"echo",
            "serialized_server_utilization_metrics",
            b"serialized_server_utilization_metrics",
        ],
    ) -> None: ...

global___PingReply = PingReply

@typing_extensions.final
class StreamingPingRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SEQUENCE_LENGTH_FIELD_NUMBER: builtins.int
    ECHO_FIELD_NUMBER: builtins.int
    sequence_length: builtins.int
    echo: builtins.str
    def __init__(
        self,
        *,
        sequence_length: builtins.int = ...,
        echo: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "echo", b"echo", "sequence_length", b"sequence_length"
        ],
    ) -> None: ...

global___StreamingPingRequest = StreamingPingRequest

@typing_extensions.final
class StreamingPingEvent(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SEQUENCE_NUMBER_FIELD_NUMBER: builtins.int
    ECHO_FIELD_NUMBER: builtins.int
    sequence_number: builtins.int
    echo: builtins.str
    def __init__(
        self,
        *,
        sequence_number: builtins.int = ...,
        echo: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "echo", b"echo", "sequence_number", b"sequence_number"
        ],
    ) -> None: ...

global___StreamingPingEvent = StreamingPingEvent

@typing_extensions.final
class GetServerIdReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERVER_ID_FIELD_NUMBER: builtins.int
    server_id: builtins.str
    def __init__(
        self,
        *,
        server_id: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["server_id", b"server_id"]
    ) -> None: ...

global___GetServerIdReply = GetServerIdReply

@typing_extensions.final
class ExecutionPlanSnapshotRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_EXECUTION_PLAN_SNAPSHOT_ARGS_FIELD_NUMBER: builtins.int
    serialized_execution_plan_snapshot_args: builtins.str
    def __init__(
        self,
        *,
        serialized_execution_plan_snapshot_args: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_execution_plan_snapshot_args", b"serialized_execution_plan_snapshot_args"
        ],
    ) -> None: ...

global___ExecutionPlanSnapshotRequest = ExecutionPlanSnapshotRequest

@typing_extensions.final
class ExecutionPlanSnapshotReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_EXECUTION_PLAN_SNAPSHOT_FIELD_NUMBER: builtins.int
    serialized_execution_plan_snapshot: builtins.str
    def __init__(
        self,
        *,
        serialized_execution_plan_snapshot: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_execution_plan_snapshot", b"serialized_execution_plan_snapshot"
        ],
    ) -> None: ...

global___ExecutionPlanSnapshotReply = ExecutionPlanSnapshotReply

@typing_extensions.final
class ExternalPartitionNamesRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_PARTITION_NAMES_ARGS_FIELD_NUMBER: builtins.int
    serialized_partition_names_args: builtins.str
    def __init__(
        self,
        *,
        serialized_partition_names_args: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_partition_names_args", b"serialized_partition_names_args"
        ],
    ) -> None: ...

global___ExternalPartitionNamesRequest = ExternalPartitionNamesRequest

@typing_extensions.final
class ExternalPartitionNamesReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_EXTERNAL_PARTITION_NAMES_OR_EXTERNAL_PARTITION_EXECUTION_ERROR_FIELD_NUMBER: (
        builtins.int
    )
    serialized_external_partition_names_or_external_partition_execution_error: builtins.str
    def __init__(
        self,
        *,
        serialized_external_partition_names_or_external_partition_execution_error: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_external_partition_names_or_external_partition_execution_error",
            b"serialized_external_partition_names_or_external_partition_execution_error",
        ],
    ) -> None: ...

global___ExternalPartitionNamesReply = ExternalPartitionNamesReply

@typing_extensions.final
class ExternalNotebookDataRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NOTEBOOK_PATH_FIELD_NUMBER: builtins.int
    notebook_path: builtins.str
    def __init__(
        self,
        *,
        notebook_path: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["notebook_path", b"notebook_path"]
    ) -> None: ...

global___ExternalNotebookDataRequest = ExternalNotebookDataRequest

@typing_extensions.final
class ExternalNotebookDataReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CONTENT_FIELD_NUMBER: builtins.int
    content: builtins.bytes
    def __init__(
        self,
        *,
        content: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["content", b"content"]) -> None: ...

global___ExternalNotebookDataReply = ExternalNotebookDataReply

@typing_extensions.final
class ExternalPartitionConfigRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_PARTITION_ARGS_FIELD_NUMBER: builtins.int
    serialized_partition_args: builtins.str
    def __init__(
        self,
        *,
        serialized_partition_args: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_partition_args", b"serialized_partition_args"
        ],
    ) -> None: ...

global___ExternalPartitionConfigRequest = ExternalPartitionConfigRequest

@typing_extensions.final
class ExternalPartitionConfigReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_EXTERNAL_PARTITION_CONFIG_OR_EXTERNAL_PARTITION_EXECUTION_ERROR_FIELD_NUMBER: (
        builtins.int
    )
    serialized_external_partition_config_or_external_partition_execution_error: builtins.str
    def __init__(
        self,
        *,
        serialized_external_partition_config_or_external_partition_execution_error: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_external_partition_config_or_external_partition_execution_error",
            b"serialized_external_partition_config_or_external_partition_execution_error",
        ],
    ) -> None: ...

global___ExternalPartitionConfigReply = ExternalPartitionConfigReply

@typing_extensions.final
class ExternalPartitionTagsRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_PARTITION_ARGS_FIELD_NUMBER: builtins.int
    serialized_partition_args: builtins.str
    def __init__(
        self,
        *,
        serialized_partition_args: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_partition_args", b"serialized_partition_args"
        ],
    ) -> None: ...

global___ExternalPartitionTagsRequest = ExternalPartitionTagsRequest

@typing_extensions.final
class ExternalPartitionTagsReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_EXTERNAL_PARTITION_TAGS_OR_EXTERNAL_PARTITION_EXECUTION_ERROR_FIELD_NUMBER: (
        builtins.int
    )
    serialized_external_partition_tags_or_external_partition_execution_error: builtins.str
    def __init__(
        self,
        *,
        serialized_external_partition_tags_or_external_partition_execution_error: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_external_partition_tags_or_external_partition_execution_error",
            b"serialized_external_partition_tags_or_external_partition_execution_error",
        ],
    ) -> None: ...

global___ExternalPartitionTagsReply = ExternalPartitionTagsReply

@typing_extensions.final
class ExternalPartitionSetExecutionParamsRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_PARTITION_SET_EXECUTION_PARAM_ARGS_FIELD_NUMBER: builtins.int
    serialized_partition_set_execution_param_args: builtins.str
    def __init__(
        self,
        *,
        serialized_partition_set_execution_param_args: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_partition_set_execution_param_args",
            b"serialized_partition_set_execution_param_args",
        ],
    ) -> None: ...

global___ExternalPartitionSetExecutionParamsRequest = ExternalPartitionSetExecutionParamsRequest

@typing_extensions.final
class ListRepositoriesRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___ListRepositoriesRequest = ListRepositoriesRequest

@typing_extensions.final
class ListRepositoriesReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_LIST_REPOSITORIES_RESPONSE_OR_ERROR_FIELD_NUMBER: builtins.int
    serialized_list_repositories_response_or_error: builtins.str
    def __init__(
        self,
        *,
        serialized_list_repositories_response_or_error: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_list_repositories_response_or_error",
            b"serialized_list_repositories_response_or_error",
        ],
    ) -> None: ...

global___ListRepositoriesReply = ListRepositoriesReply

@typing_extensions.final
class ExternalPipelineSubsetSnapshotRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_PIPELINE_SUBSET_SNAPSHOT_ARGS_FIELD_NUMBER: builtins.int
    serialized_pipeline_subset_snapshot_args: builtins.str
    def __init__(
        self,
        *,
        serialized_pipeline_subset_snapshot_args: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_pipeline_subset_snapshot_args", b"serialized_pipeline_subset_snapshot_args"
        ],
    ) -> None: ...

global___ExternalPipelineSubsetSnapshotRequest = ExternalPipelineSubsetSnapshotRequest

@typing_extensions.final
class ExternalPipelineSubsetSnapshotReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_EXTERNAL_PIPELINE_SUBSET_RESULT_FIELD_NUMBER: builtins.int
    serialized_external_pipeline_subset_result: builtins.str
    def __init__(
        self,
        *,
        serialized_external_pipeline_subset_result: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_external_pipeline_subset_result",
            b"serialized_external_pipeline_subset_result",
        ],
    ) -> None: ...

global___ExternalPipelineSubsetSnapshotReply = ExternalPipelineSubsetSnapshotReply

@typing_extensions.final
class ExternalRepositoryRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_REPOSITORY_PYTHON_ORIGIN_FIELD_NUMBER: builtins.int
    DEFER_SNAPSHOTS_FIELD_NUMBER: builtins.int
    serialized_repository_python_origin: builtins.str
    defer_snapshots: builtins.bool
    def __init__(
        self,
        *,
        serialized_repository_python_origin: builtins.str = ...,
        defer_snapshots: builtins.bool = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "defer_snapshots",
            b"defer_snapshots",
            "serialized_repository_python_origin",
            b"serialized_repository_python_origin",
        ],
    ) -> None: ...

global___ExternalRepositoryRequest = ExternalRepositoryRequest

@typing_extensions.final
class ExternalRepositoryReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_EXTERNAL_REPOSITORY_DATA_FIELD_NUMBER: builtins.int
    serialized_external_repository_data: builtins.str
    def __init__(
        self,
        *,
        serialized_external_repository_data: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_external_repository_data", b"serialized_external_repository_data"
        ],
    ) -> None: ...

global___ExternalRepositoryReply = ExternalRepositoryReply

@typing_extensions.final
class StreamingExternalRepositoryEvent(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SEQUENCE_NUMBER_FIELD_NUMBER: builtins.int
    SERIALIZED_EXTERNAL_REPOSITORY_CHUNK_FIELD_NUMBER: builtins.int
    sequence_number: builtins.int
    serialized_external_repository_chunk: builtins.str
    def __init__(
        self,
        *,
        sequence_number: builtins.int = ...,
        serialized_external_repository_chunk: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "sequence_number",
            b"sequence_number",
            "serialized_external_repository_chunk",
            b"serialized_external_repository_chunk",
        ],
    ) -> None: ...

global___StreamingExternalRepositoryEvent = StreamingExternalRepositoryEvent

@typing_extensions.final
class ExternalScheduleExecutionRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_EXTERNAL_SCHEDULE_EXECUTION_ARGS_FIELD_NUMBER: builtins.int
    serialized_external_schedule_execution_args: builtins.str
    def __init__(
        self,
        *,
        serialized_external_schedule_execution_args: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_external_schedule_execution_args",
            b"serialized_external_schedule_execution_args",
        ],
    ) -> None: ...

global___ExternalScheduleExecutionRequest = ExternalScheduleExecutionRequest

@typing_extensions.final
class ExternalSensorExecutionRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_EXTERNAL_SENSOR_EXECUTION_ARGS_FIELD_NUMBER: builtins.int
    serialized_external_sensor_execution_args: builtins.str
    def __init__(
        self,
        *,
        serialized_external_sensor_execution_args: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_external_sensor_execution_args",
            b"serialized_external_sensor_execution_args",
        ],
    ) -> None: ...

global___ExternalSensorExecutionRequest = ExternalSensorExecutionRequest

@typing_extensions.final
class StreamingChunkEvent(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SEQUENCE_NUMBER_FIELD_NUMBER: builtins.int
    SERIALIZED_CHUNK_FIELD_NUMBER: builtins.int
    sequence_number: builtins.int
    serialized_chunk: builtins.str
    def __init__(
        self,
        *,
        sequence_number: builtins.int = ...,
        serialized_chunk: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "sequence_number", b"sequence_number", "serialized_chunk", b"serialized_chunk"
        ],
    ) -> None: ...

global___StreamingChunkEvent = StreamingChunkEvent

@typing_extensions.final
class ShutdownServerReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_SHUTDOWN_SERVER_RESULT_FIELD_NUMBER: builtins.int
    serialized_shutdown_server_result: builtins.str
    def __init__(
        self,
        *,
        serialized_shutdown_server_result: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_shutdown_server_result", b"serialized_shutdown_server_result"
        ],
    ) -> None: ...

global___ShutdownServerReply = ShutdownServerReply

@typing_extensions.final
class CancelExecutionRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_CANCEL_EXECUTION_REQUEST_FIELD_NUMBER: builtins.int
    serialized_cancel_execution_request: builtins.str
    def __init__(
        self,
        *,
        serialized_cancel_execution_request: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_cancel_execution_request", b"serialized_cancel_execution_request"
        ],
    ) -> None: ...

global___CancelExecutionRequest = CancelExecutionRequest

@typing_extensions.final
class CancelExecutionReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_CANCEL_EXECUTION_RESULT_FIELD_NUMBER: builtins.int
    serialized_cancel_execution_result: builtins.str
    def __init__(
        self,
        *,
        serialized_cancel_execution_result: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_cancel_execution_result", b"serialized_cancel_execution_result"
        ],
    ) -> None: ...

global___CancelExecutionReply = CancelExecutionReply

@typing_extensions.final
class CanCancelExecutionRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_CAN_CANCEL_EXECUTION_REQUEST_FIELD_NUMBER: builtins.int
    serialized_can_cancel_execution_request: builtins.str
    def __init__(
        self,
        *,
        serialized_can_cancel_execution_request: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_can_cancel_execution_request", b"serialized_can_cancel_execution_request"
        ],
    ) -> None: ...

global___CanCancelExecutionRequest = CanCancelExecutionRequest

@typing_extensions.final
class CanCancelExecutionReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_CAN_CANCEL_EXECUTION_RESULT_FIELD_NUMBER: builtins.int
    serialized_can_cancel_execution_result: builtins.str
    def __init__(
        self,
        *,
        serialized_can_cancel_execution_result: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_can_cancel_execution_result", b"serialized_can_cancel_execution_result"
        ],
    ) -> None: ...

global___CanCancelExecutionReply = CanCancelExecutionReply

@typing_extensions.final
class StartRunRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_EXECUTE_RUN_ARGS_FIELD_NUMBER: builtins.int
    serialized_execute_run_args: builtins.str
    def __init__(
        self,
        *,
        serialized_execute_run_args: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_execute_run_args", b"serialized_execute_run_args"
        ],
    ) -> None: ...

global___StartRunRequest = StartRunRequest

@typing_extensions.final
class StartRunReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_START_RUN_RESULT_FIELD_NUMBER: builtins.int
    serialized_start_run_result: builtins.str
    def __init__(
        self,
        *,
        serialized_start_run_result: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_start_run_result", b"serialized_start_run_result"
        ],
    ) -> None: ...

global___StartRunReply = StartRunReply

@typing_extensions.final
class GetCurrentImageReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_CURRENT_IMAGE_FIELD_NUMBER: builtins.int
    serialized_current_image: builtins.str
    def __init__(
        self,
        *,
        serialized_current_image: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_current_image", b"serialized_current_image"
        ],
    ) -> None: ...

global___GetCurrentImageReply = GetCurrentImageReply

@typing_extensions.final
class GetCurrentRunsReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_CURRENT_RUNS_FIELD_NUMBER: builtins.int
    serialized_current_runs: builtins.str
    def __init__(
        self,
        *,
        serialized_current_runs: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_current_runs", b"serialized_current_runs"
        ],
    ) -> None: ...

global___GetCurrentRunsReply = GetCurrentRunsReply

@typing_extensions.final
class ExternalJobRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_REPOSITORY_ORIGIN_FIELD_NUMBER: builtins.int
    JOB_NAME_FIELD_NUMBER: builtins.int
    serialized_repository_origin: builtins.str
    job_name: builtins.str
    def __init__(
        self,
        *,
        serialized_repository_origin: builtins.str = ...,
        job_name: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "job_name", b"job_name", "serialized_repository_origin", b"serialized_repository_origin"
        ],
    ) -> None: ...

global___ExternalJobRequest = ExternalJobRequest

@typing_extensions.final
class ExternalJobReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_JOB_DATA_FIELD_NUMBER: builtins.int
    SERIALIZED_ERROR_FIELD_NUMBER: builtins.int
    serialized_job_data: builtins.str
    serialized_error: builtins.str
    def __init__(
        self,
        *,
        serialized_job_data: builtins.str = ...,
        serialized_error: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_error", b"serialized_error", "serialized_job_data", b"serialized_job_data"
        ],
    ) -> None: ...

global___ExternalJobReply = ExternalJobReply

@typing_extensions.final
class ExternalScheduleExecutionReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_SCHEDULE_RESULT_FIELD_NUMBER: builtins.int
    serialized_schedule_result: builtins.str
    def __init__(
        self,
        *,
        serialized_schedule_result: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_schedule_result", b"serialized_schedule_result"
        ],
    ) -> None: ...

global___ExternalScheduleExecutionReply = ExternalScheduleExecutionReply

@typing_extensions.final
class ExternalSensorExecutionReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_SENSOR_RESULT_FIELD_NUMBER: builtins.int
    serialized_sensor_result: builtins.str
    def __init__(
        self,
        *,
        serialized_sensor_result: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "serialized_sensor_result", b"serialized_sensor_result"
        ],
    ) -> None: ...

global___ExternalSensorExecutionReply = ExternalSensorExecutionReply

@typing_extensions.final
class ReloadCodeRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___ReloadCodeRequest = ReloadCodeRequest

@typing_extensions.final
class ReloadCodeReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERIALIZED_ERROR_FIELD_NUMBER: builtins.int
    serialized_error: builtins.str
    def __init__(
        self,
        *,
        serialized_error: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["serialized_error", b"serialized_error"]
    ) -> None: ...

global___ReloadCodeReply = ReloadCodeReply



import re
from collections.abc import Mapping
from enum import Enum
from functools import cached_property
from typing import Any, Optional, cast

from pydantic import BaseModel, ConfigDict
from typing_extensions import TypeVar

import dagster._check as check
from dagster import (
    Field as DagsterField,
    Shape,
)
from dagster._config.field_utils import EnvVar, IntEnvVar, Permissive
from dagster._config.pythonic_config.attach_other_object_to_context import (
    IAttachDifferentObjectToOpContext as IAttachDifferentObjectToOpContext,
)
from dagster._config.pythonic_config.conversion_utils import (
    _convert_pydantic_field,
    safe_is_subclass,
)
from dagster._config.pythonic_config.type_check_utils import is_literal
from dagster._config.pythonic_config.typing_utils import BaseConfigMeta
from dagster._core.definitions.definition_config_schema import DefinitionConfigSchema
from dagster._core.errors import (
    DagsterInvalidConfigDefinitionError,
    DagsterInvalidDefinitionError,
    DagsterInvalidInvocationError,
    DagsterInvalidPythonicConfigDefinitionError,
)
from dagster._model.pydantic_compat_layer import (
    ModelFieldCompat,
    PydanticUndefined,
    model_config,
    model_fields,
)
from dagster._utils.cached_method import CACHED_METHOD_CACHE_FIELD

INTERNAL_MARKER = "__internal__"


def _is_field_internal(name: str) -> bool:
    return name.endswith(INTERNAL_MARKER)


# ensure that this ends with the internal marker so we can do a single check
assert CACHED_METHOD_CACHE_FIELD.endswith(INTERNAL_MARKER)


def _is_frozen_pydantic_error(e: Exception) -> bool:
    """Parses an error to determine if it is a Pydantic error indicating that the instance
    is immutable. We use this to attach a more helpful error message.
    """
    return "Instance is frozen" in str(  # Pydantic 2.x error
        e
    ) or "is immutable and does not support item assignment" in str(  # Pydantic 1.x error
        e
    )


class MakeConfigCacheable(BaseModel):
    """This class centralizes and implements all the chicanery we need in order
    to support caching decorators. If we decide this is a bad idea we can remove it
    all in one go.
    """

    # - Frozen, to avoid complexity caused by mutation.
    # - arbitrary_types_allowed, to allow non-model class params to be validated with isinstance.
    # - Avoid pydantic reading a cached property class as part of the schema.
    model_config = ConfigDict(
        frozen=True, arbitrary_types_allowed=True, ignored_types=(cached_property,)
    )

    def __setattr__(self, name: str, value: Any):
        from dagster._config.pythonic_config.resource import ConfigurableResourceFactory

        # This is a hack to allow us to set attributes on the class that are not part of the
        # config schema. Pydantic will normally raise an error if you try to set an attribute
        # that is not part of the schema.

        if _is_field_internal(name):
            object.__setattr__(self, name, value)
            return

        try:
            return super().__setattr__(name, value)
        except (TypeError, ValueError) as e:
            clsname = self.__class__.__name__
            if _is_frozen_pydantic_error(e):
                if isinstance(self, ConfigurableResourceFactory):
                    raise DagsterInvalidInvocationError(
                        f"'{clsname}' is a Pythonic resource and does not support item assignment,"
                        " as it inherits from 'pydantic.BaseModel' with frozen=True. If trying to"
                        " maintain state on this resource, consider building a separate, stateful"
                        " client class, and provide a method on the resource to construct and"
                        " return the stateful client."
                    ) from e
                else:
                    raise DagsterInvalidInvocationError(
                        f"'{clsname}' is a Pythonic config class and does not support item"
                        " assignment, as it inherits from 'pydantic.BaseModel' with frozen=True."
                    ) from e
            elif "object has no field" in str(e):
                field_name = check.not_none(
                    re.search(r"object has no field \"(.*)\"", str(e))
                ).group(1)
                if isinstance(self, ConfigurableResourceFactory):
                    raise DagsterInvalidInvocationError(
                        f"'{clsname}' is a Pythonic resource and does not support manipulating"
                        f" undeclared attribute '{field_name}' as it inherits from"
                        " 'pydantic.BaseModel' without extra=\"allow\". If trying to maintain"
                        " state on this resource, consider building a separate, stateful client"
                        " class, and provide a method on the resource to construct and return the"
                        " stateful client."
                    ) from e
                else:
                    raise DagsterInvalidInvocationError(
                        f"'{clsname}' is a Pythonic config class and does not support manipulating"
                        f" undeclared attribute '{field_name}' as it inherits from"
                        " 'pydantic.BaseModel' without extra=\"allow\"."
                    ) from e
            else:
                raise


T = TypeVar("T")


def ensure_env_vars_set_post_init(set_value: T, input_value: Any) -> T:
    """Pydantic 2.x utility. Ensures that Pydantic field values are set to the appropriate
    EnvVar or IntEnvVar objects post-model-instantiation, since Pydantic 2.x will cast
    EnvVar or IntEnvVar values to raw strings or ints as part of the model instantiation process.
    """
    if isinstance(set_value, dict) and isinstance(input_value, dict):
        for key, value in input_value.items():
            if isinstance(value, (EnvVar, IntEnvVar)):
                set_value[key] = value
            elif isinstance(value, dict):
                set_value[key] = ensure_env_vars_set_post_init(set_value.get(key) or {}, value)
            elif isinstance(value, list):
                set_value[key] = ensure_env_vars_set_post_init(set_value.get(key) or [], value)
    if isinstance(set_value, list) and isinstance(input_value, list):
        for i in range(len(set_value)):
            value = input_value[i]
            if isinstance(value, (EnvVar, IntEnvVar)):
                set_value[i] = value
            elif isinstance(value, (dict, list)):
                set_value[i] = ensure_env_vars_set_post_init(set_value[i], value)

    return set_value


class Config(MakeConfigCacheable, metaclass=BaseConfigMeta):
    """Base class for Dagster configuration models, used to specify config schema for
    ops and assets. Subclasses :py:class:`pydantic.BaseModel`.

    Example definition:

    .. code-block:: python

        from pydantic import Field

        class MyAssetConfig(Config):
            my_str: str = "my_default_string"
            my_int_list: List[int]
            my_bool_with_metadata: bool = Field(default=False, description="A bool field")


    Example usage:

    .. code-block:: python

        @asset
        def asset_with_config(config: MyAssetConfig):
            assert config.my_str == "my_default_string"
            assert config.my_int_list == [1, 2, 3]
            assert config.my_bool_with_metadata == False

        asset_with_config(MyAssetConfig(my_int_list=[1, 2, 3], my_bool_with_metadata=True))

    """

    def __init__(self, **config_dict) -> None:
        """This constructor is overridden to handle any remapping of raw config dicts to
        the appropriate config classes. For example, discriminated unions are represented
        in Dagster config as dicts with a single key, which is the discriminator value.
        """
        # In order to respect aliases on pydantic fields, we need to keep track of
        # both the the field_key which is they key for the field on the pydantic model
        # and the config_key which is the post alias resolution name which is
        # the key for that field in the incoming config_dict
        modified_data_by_config_key = {}
        field_info_by_config_key = {
            field.alias if field.alias else field_key: (field_key, field)
            for field_key, field in model_fields(self).items()
        }
        for config_key, value in config_dict.items():
            field_info = field_info_by_config_key.get(config_key)
            field = None
            field_key = config_key
            if field_info:
                field_key, field = field_info

            if field and field.discriminator:
                nested_dict = value

                discriminator_key = check.not_none(field.discriminator)
                if isinstance(value, Config):
                    nested_dict = _discriminated_union_config_dict_to_selector_config_dict(
                        discriminator_key,
                        value._get_non_default_public_field_values(),  # noqa: SLF001
                    )

                nested_items = list(check.is_dict(nested_dict).items())
                check.invariant(
                    len(nested_items) == 1,
                    "Discriminated union must have exactly one key",
                )
                discriminated_value, nested_values = nested_items[0]

                modified_data_by_config_key[config_key] = {
                    **nested_values,
                    discriminator_key: discriminated_value,
                }

            # If the passed value matches the name of an expected Enum value, convert it to the value
            elif (
                field
                and safe_is_subclass(field.annotation, Enum)
                and value in field.annotation.__members__
                and value not in [member.value for member in field.annotation]  # type: ignore
            ):
                modified_data_by_config_key[config_key] = field.annotation.__members__[value].value
            elif field and safe_is_subclass(field.annotation, Config) and isinstance(value, dict):
                modified_data_by_config_key[field_key] = (
                    field.annotation._get_non_default_public_field_values_cls(  # noqa: SLF001
                        value
                    )
                )
            else:
                modified_data_by_config_key[config_key] = value

        for field_key, field in model_fields(self).items():
            config_key = field.alias if field.alias else field_key
            if field.is_required() and config_key not in modified_data_by_config_key:
                modified_data_by_config_key[config_key] = (
                    field.default if field.default != PydanticUndefined else None
                )

        super().__init__(**modified_data_by_config_key)

        modified_data_by_field_key = {}
        for config_key, value in modified_data_by_config_key.items():
            field_info = field_info_by_config_key.get(config_key)
            field_key = field_info[0] if field_info else config_key
            modified_data_by_field_key[field_key] = value

        self.__dict__ = ensure_env_vars_set_post_init(self.__dict__, modified_data_by_field_key)

    def _convert_to_config_dictionary(self) -> Mapping[str, Any]:
        """Converts this Config object to a Dagster config dictionary, in the same format as the dictionary
        accepted as run config or as YAML in the launchpad.

        Inner fields are recursively converted to dictionaries, meaning nested config objects
        or EnvVars will be converted to the appropriate dictionary representation.
        """
        public_fields = self._get_non_default_public_field_values(keep_if_not_none=True)
        return {
            k: _config_value_to_dict_representation(model_fields(self).get(k), v)
            for k, v in public_fields.items()
        }

    @classmethod
    def _get_non_default_public_field_values_cls(
        cls, items: dict[str, Any], keep_if_not_none: bool = False
    ) -> Mapping[str, Any]:
        """Returns a dictionary representation of this config object,
        ignoring any private fields, and any defaulted fields which are equal to the default value
        (unless `keep_if_not_none`, in which case any non-`None` defaulted fields are kept).

        Inner fields are returned as-is in the dictionary,
        meaning any nested config objects will be returned as config objects, not dictionaries.
        """
        output = {}
        for key, value in items.items():
            if _is_field_internal(key):
                continue
            field = model_fields(cls).get(key)

            if field:
                keep = keep_if_not_none and value is not None
                if (
                    not keep
                    and not is_literal(field.annotation)
                    and not safe_is_subclass(field.annotation, Enum)
                    and value == field.default
                ):
                    continue

                resolved_field_name = field.alias or key
                output[resolved_field_name] = value
            else:
                output[key] = value
        return output

    def _get_non_default_public_field_values(
        self, keep_if_not_none: bool = False
    ) -> Mapping[str, Any]:
        return self.__class__._get_non_default_public_field_values_cls(dict(self), keep_if_not_none)  # noqa: SLF001

    @classmethod
    def to_config_schema(cls) -> DefinitionConfigSchema:
        """Converts the config structure represented by this class into a DefinitionConfigSchema."""
        return DefinitionConfigSchema(infer_schema_from_config_class(cls))

    @classmethod
    def to_fields_dict(cls) -> dict[str, DagsterField]:
        """Converts the config structure represented by this class into a dictionary of dagster.Fields.
        This is useful when interacting with legacy code that expects a dictionary of fields but you
        want the source of truth to be a config class.
        """
        return cast(Shape, cls.to_config_schema().as_field().config_type).fields  # pyright: ignore[reportReturnType]


def _discriminated_union_config_dict_to_selector_config_dict(
    discriminator_key: str, config_dict: Mapping[str, Any]
):
    """Remaps a config dictionary which is a member of a discriminated union to
    the appropriate structure for a Dagster config selector.

    A discriminated union with key "my_key" and value "my_value" will be represented
    as {"my_key": "my_value", "my_field": "my_field_value"}. When converted to a selector,
    this should be represented as {"my_value": {"my_field": "my_field_value"}}.
    """
    updated_dict = dict(config_dict)
    discriminator_value = updated_dict.pop(discriminator_key)
    wrapped_dict = {discriminator_value: updated_dict}
    return wrapped_dict


def _config_value_to_dict_representation(field: Optional[ModelFieldCompat], value: Any):
    """Converts a config value to a dictionary representation. If a field is provided, it will be used
    to determine the appropriate dictionary representation in the case of discriminated unions.
    """
    from dagster._config.field_utils import env_var_to_config_dict, is_dagster_env_var

    if isinstance(value, dict):
        return {k: _config_value_to_dict_representation(None, v) for k, v in value.items()}
    elif isinstance(value, list):
        return [_config_value_to_dict_representation(None, v) for v in value]
    elif is_dagster_env_var(value):
        return env_var_to_config_dict(value)
    if isinstance(value, Config):
        if field and field.discriminator:
            return {
                k: v
                for k, v in _discriminated_union_config_dict_to_selector_config_dict(
                    field.discriminator,
                    value._convert_to_config_dictionary(),  # noqa: SLF001
                ).items()
            }
        else:
            return {k: v for k, v in value._convert_to_config_dictionary().items()}  # noqa: SLF001
    elif isinstance(value, Enum):
        return value.name

    return value


class PermissiveConfig(Config):
    """Subclass of :py:class:`Config` that allows arbitrary extra fields. This is useful for
    config classes which may have open-ended inputs.

    Example definition:

    .. code-block:: python

        class MyPermissiveOpConfig(PermissiveConfig):
            my_explicit_parameter: bool
            my_other_explicit_parameter: str


    Example usage:

    .. code-block:: python

        @op
        def op_with_config(config: MyPermissiveOpConfig):
            assert config.my_explicit_parameter == True
            assert config.my_other_explicit_parameter == "foo"
            assert config.dict().get("my_implicit_parameter") == "bar"

        op_with_config(
            MyPermissiveOpConfig(
                my_explicit_parameter=True,
                my_other_explicit_parameter="foo",
                my_implicit_parameter="bar"
            )
        )

    """

    # Pydantic config for this class
    # Cannot use kwargs for base class as this is not support for pydantic<1.8
    model_config = ConfigDict(extra="allow")


def infer_schema_from_config_class(
    model_cls: type["Config"],
    description: Optional[str] = None,
    fields_to_omit: Optional[set[str]] = None,
    default: Optional[Any] = None,
) -> DagsterField:
    from dagster._config.pythonic_config.config import Config
    from dagster._config.pythonic_config.resource import (
        ConfigurableResourceFactory,
        _is_annotated_as_resource_type,
    )

    """Parses a structured config class and returns a corresponding Dagster config Field."""
    fields_to_omit = fields_to_omit or set()

    check.param_invariant(
        safe_is_subclass(model_cls, Config),
        "Config type annotation must inherit from dagster.Config",
    )

    fields: dict[str, DagsterField] = {}
    for key, pydantic_field_info in model_fields(model_cls).items():
        if _is_annotated_as_resource_type(
            pydantic_field_info.annotation, pydantic_field_info.metadata
        ):
            continue

        resolved_field_name = pydantic_field_info.alias if pydantic_field_info.alias else key
        if key not in fields_to_omit:
            if isinstance(pydantic_field_info.default, DagsterField):
                raise DagsterInvalidDefinitionError(
                    "Using 'dagster.Field' is not supported within a Pythonic config or resource"
                    " definition. 'dagster.Field' should only be used in legacy Dagster config"
                    " schemas. Did you mean to use 'pydantic.Field' instead?"
                )
            field_default = default.get(resolved_field_name) if isinstance(default, dict) else None
            try:
                fields[resolved_field_name] = _convert_pydantic_field(
                    pydantic_field_info, default=field_default
                )
            except DagsterInvalidConfigDefinitionError as e:
                raise DagsterInvalidPythonicConfigDefinitionError(
                    config_class=model_cls,
                    field_name=key,
                    invalid_type=e.current_value,
                    is_resource=model_cls is not None
                    and safe_is_subclass(model_cls, ConfigurableResourceFactory),
                )

    shape_cls = Permissive if model_config(model_cls).get("extra") == "allow" else Shape

    docstring = model_cls.__doc__.strip() if model_cls.__doc__ else None

    return DagsterField(config=shape_cls(fields), description=description or docstring)



import sys
from collections.abc import Iterator, Mapping, Sequence
from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import datetime
from functools import cached_property
from queue import Queue
from typing import TYPE_CHECKING, Any, Literal, Optional, TypedDict, Union, cast

from dagster_pipes import (
    DAGSTER_PIPES_CONTEXT_ENV_VAR,
    DAGSTER_PIPES_MESSAGES_ENV_VAR,
    PIPES_METADATA_TYPE_INFER,
    Method,
    PipesContextData,
    PipesDataProvenance,
    PipesException,
    PipesExtras,
    PipesMessage,
    PipesMetadataType,
    PipesMetadataValue,
    PipesOpenedData,
    PipesParams,
    PipesTimeWindow,
    _env_var_to_cli_argument,
    encode_param,
)
from typing_extensions import TypeAlias

import dagster._check as check
from dagster import DagsterEvent
from dagster._annotations import public
from dagster._core.definitions.asset_check_result import AssetCheckResult
from dagster._core.definitions.asset_check_spec import AssetCheckSeverity
from dagster._core.definitions.data_version import DataProvenance, DataVersion
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.metadata import MetadataValue, normalize_metadata_value
from dagster._core.definitions.metadata.table import (
    TableColumn,
    TableColumnConstraints,
    TableColumnDep,
    TableColumnLineage,
    TableRecord,
    TableSchema,
)
from dagster._core.definitions.partition_key_range import PartitionKeyRange
from dagster._core.definitions.result import MaterializeResult
from dagster._core.definitions.time_window_partitions import (
    TimeWindow,
    has_one_dimension_time_window_partitioning,
)
from dagster._core.errors import DagsterInvariantViolationError, DagsterPipesExecutionError
from dagster._core.events import EngineEventData
from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
from dagster._core.execution.context.invocation import BaseDirectExecutionContext
from dagster._core.execution.context.op_execution_context import OpExecutionContext
from dagster._utils.error import (
    ExceptionInfo,
    SerializableErrorInfo,
    serializable_error_info_from_exc_info,
)

if TYPE_CHECKING:
    from dagster._core.pipes.client import PipesMessageReader


PipesExecutionResult: TypeAlias = Union[MaterializeResult, AssetCheckResult]


class PipesLaunchedData(TypedDict):
    """Payload generated in the orchestration process after external process startup
    containing arbitrary information about the external process.
    """

    extras: Mapping[str, Any]


class PipesMessageHandler:
    """Class to process :py:obj:`PipesMessage` objects received from a pipes process.

    Args:
        context (Union[OpExecutionContext, AssetExecutionContext]): The context for the executing op/asset.
        message_reader (PipesMessageReader): The message reader used to read messages from the
            external process.
    """

    # In the future it may make sense to merge PipesMessageReader and PipesMessageHandler, or
    # otherwise adjust their relationship. The current interaction between the two is a bit awkward,
    # but it would also be awkward to have a monolith that users extend.
    def __init__(
        self,
        context: Union[OpExecutionContext, AssetExecutionContext],
        message_reader: "PipesMessageReader",
    ) -> None:
        self._context = context
        self._message_reader = message_reader
        # Queue is thread-safe
        self._result_queue: Queue[PipesExecutionResult] = Queue()
        self._extra_msg_queue: Queue[Any] = Queue()
        # Only read by the main thread after all messages are handled, so no need for a lock
        self._received_opened_msg = False
        self._messages_include_stdio_logs = False
        self._received_closed_msg = False
        self._opened_payload: Optional[PipesOpenedData] = None

    @contextmanager
    def handle_messages(self) -> Iterator[PipesParams]:
        with self._message_reader.read_messages(self) as params:
            yield params

    def get_reported_results(self) -> Sequence[PipesExecutionResult]:
        return tuple(self._result_queue.queue)

    def get_custom_messages(self) -> Sequence[Any]:
        return tuple(self._extra_msg_queue.queue)

    @property
    def received_opened_message(self) -> bool:
        return self._received_opened_msg

    @property
    def received_closed_message(self) -> bool:
        return self._received_closed_msg

    def _resolve_metadata(
        self, metadata: Mapping[str, PipesMetadataValue]
    ) -> Mapping[str, MetadataValue]:
        return {
            k: self._resolve_metadata_value(v["raw_value"], v["type"]) for k, v in metadata.items()
        }

    @staticmethod
    def _resolve_metadata_value(value: Any, metadata_type: PipesMetadataType) -> MetadataValue:
        if metadata_type == PIPES_METADATA_TYPE_INFER:
            return normalize_metadata_value(value)
        elif metadata_type == "text":
            return MetadataValue.text(value)
        elif metadata_type == "url":
            return MetadataValue.url(value)
        elif metadata_type == "path":
            return MetadataValue.path(value)
        elif metadata_type == "notebook":
            return MetadataValue.notebook(value)
        elif metadata_type == "json":
            return MetadataValue.json(value)
        elif metadata_type == "md":
            return MetadataValue.md(value)
        elif metadata_type == "float":
            return MetadataValue.float(value)
        elif metadata_type == "int":
            return MetadataValue.int(value)
        elif metadata_type == "bool":
            return MetadataValue.bool(value)
        elif metadata_type == "dagster_run":
            return MetadataValue.dagster_run(value)
        elif metadata_type == "asset":
            return MetadataValue.asset(AssetKey.from_user_string(value))
        elif metadata_type == "table":
            value = check.mapping_param(value, "table_value", key_type=str)
            return MetadataValue.table(
                records=[TableRecord(record) for record in value["records"]],
                schema=TableSchema(
                    columns=[
                        TableColumn(
                            name=column["name"],
                            type=column["type"],
                            description=column.get("description"),
                            tags=column.get("tags"),
                            constraints=TableColumnConstraints(**column["constraints"])
                            if column.get("constraints")
                            else None,
                        )
                        for column in value["schema"]
                    ]
                ),
            )
        elif metadata_type == "table_schema":
            value = check.mapping_param(value, "table_schema_value", key_type=str)
            return MetadataValue.table_schema(
                schema=TableSchema(
                    columns=[
                        TableColumn(
                            name=column["name"],
                            type=column["type"],
                            description=column.get("description"),
                            tags=column.get("tags"),
                            constraints=TableColumnConstraints(**column["constraints"])
                            if column.get("constraints")
                            else None,
                        )
                        for column in value["columns"]
                    ]
                )
            )
        elif metadata_type == "table_column_lineage":
            value = check.mapping_param(value, "table_column_value", key_type=str)
            return MetadataValue.column_lineage(
                lineage=TableColumnLineage(
                    deps_by_column={
                        column: [TableColumnDep(**dep) for dep in deps]
                        for column, deps in value["deps_by_column"].items()
                    }
                )
            )
        elif metadata_type == "timestamp":
            return MetadataValue.timestamp(float(check.numeric_param(value, "timestamp")))
        elif metadata_type == "null":
            return MetadataValue.null()
        else:
            check.failed(f"Unexpected metadata type {metadata_type}")

    # Type ignores because we currently validate in individual handlers
    def handle_message(self, message: PipesMessage) -> None:
        if self._received_closed_msg:
            self._context.log.warning(
                f"[pipes] unexpected message received after closed: `{message}`"
            )

        method = cast(Method, message["method"])
        if method == "opened":
            self._handle_opened(message["params"])  # type: ignore
        elif method == "closed":
            self._handle_closed(message["params"])
        elif method == "report_asset_materialization":
            self._handle_report_asset_materialization(**message["params"])  # type: ignore
        elif method == "report_asset_check":
            self._handle_report_asset_check(**message["params"])  # type: ignore
        elif method == "log":
            self._handle_log(**message["params"])  # type: ignore
        elif method == "log_external_stream":
            self._handle_log_external_stream(**message["params"])  # type: ignore
        elif method == "report_custom_message":
            self._handle_extra_message(**message["params"])  # type: ignore
        else:
            raise DagsterPipesExecutionError(f"Unknown message method: {message['method']}")

    def _handle_opened(self, opened_payload: PipesOpenedData) -> None:
        self._received_opened_msg = True
        self._context.log.info("[pipes] external process successfully opened dagster pipes.")
        self._message_reader.on_opened(opened_payload)

    def _handle_closed(self, params: Optional[Mapping[str, Any]]) -> None:
        self._received_closed_msg = True
        if params and "exception" in params:
            err_info = _ser_err_from_pipes_exc(params["exception"])
            # report as an engine event to provide structured exception data
            DagsterEvent.engine_event(
                self._context.get_step_execution_context(),
                "[pipes] external process pipes closed with exception",
                EngineEventData(error=err_info),
            )

    def _handle_report_asset_materialization(
        self,
        asset_key: str,
        metadata: Optional[Mapping[str, PipesMetadataValue]],
        data_version: Optional[str],
    ) -> None:
        check.str_param(asset_key, "asset_key")
        check.opt_str_param(data_version, "data_version")
        metadata = check.opt_mapping_param(metadata, "metadata", key_type=str)
        resolved_asset_key = AssetKey.from_escaped_user_string(asset_key)
        resolved_metadata = self._resolve_metadata(metadata)
        resolved_data_version = None if data_version is None else DataVersion(data_version)
        result = MaterializeResult(
            asset_key=resolved_asset_key,
            metadata=resolved_metadata,
            data_version=resolved_data_version,
        )
        self._result_queue.put(result)

    def _handle_report_asset_check(
        self,
        asset_key: str,
        check_name: str,
        passed: bool,
        severity: str,
        metadata: Mapping[str, PipesMetadataValue],
    ) -> None:
        check.str_param(asset_key, "asset_key")
        check.str_param(check_name, "check_name")
        check.bool_param(passed, "passed")
        check.literal_param(severity, "severity", [x.value for x in AssetCheckSeverity])
        metadata = check.opt_mapping_param(metadata, "metadata", key_type=str)
        resolved_asset_key = AssetKey.from_escaped_user_string(asset_key)
        resolved_metadata = self._resolve_metadata(metadata)
        resolved_severity = AssetCheckSeverity(severity)
        result = AssetCheckResult(
            asset_key=resolved_asset_key,
            check_name=check_name,
            passed=passed,
            severity=resolved_severity,
            metadata=resolved_metadata,
        )
        self._result_queue.put(result)

    def _handle_log(self, message: str, level: str = "info") -> None:
        check.str_param(message, "message")
        self._context.log.log(level, message)

    def _handle_log_external_stream(
        self, stream: Literal["stdout", "stderr"], text: str, extras: Optional[PipesExtras] = None
    ):
        if stream == "stdout":
            sys.stdout.write(text)
        elif stream == "stderr":
            sys.stderr.write(text)
        else:
            raise DagsterInvariantViolationError(f"Unexpected stream: {stream}")

    def _handle_extra_message(self, payload: Any):
        self._extra_msg_queue.put(payload)

    def report_pipes_framework_exception(self, origin: str, exc_info: ExceptionInfo):
        # use an engine event to provide structured exception, this gives us an event with
        # * the context of where the exception happened ([pipes]...)
        # * the exception class, message, and stack trace as well as any chained exception
        DagsterEvent.engine_event(
            self._context.get_step_execution_context(),
            f"[pipes] framework exception occurred in {origin}",
            EngineEventData(
                error=serializable_error_info_from_exc_info(exc_info),
            ),
        )

    def on_launched(self, launched_payload: PipesLaunchedData) -> None:
        """Hook that is called if `PipesSession.report_launched()` is called."""
        self._context.log.info("[pipes] additional launch_payload reported")
        self._message_reader.on_launched(launched_payload)


@dataclass
class PipesSession:
    """Object representing a pipes session.

    A pipes session is defined by a pair of :py:class:`PipesContextInjector` and
    :py:class:`PipesMessageReader` objects. At the opening of the session, the context injector
    writes context data to an externally accessible location, and the message reader starts
    monitoring an externally accessible location. These locations are encoded in parameters stored
    on a `PipesSession` object.

    During the session, an external process should be started and the parameters injected into its
    environment. The typical way to do this is to call :py:meth:`PipesSession.get_bootstrap_env_vars`
    and pass the result as environment variables.

    During execution, results (e.g. asset materializations) are reported by the external process and
    buffered on the `PipesSession` object. The buffer can periodically be cleared and yielded to
    Dagster machinery by calling `yield from PipesSession.get_results()`.

    When the external process exits, the session can be closed. Closing consists of handling any
    unprocessed messages written by the external process and cleaning up any resources used for
    context injection and message reading.

    Args:
        context_data (PipesContextData): The context for the executing op/asset.
        message_handler (PipesMessageHandler): The message handler to use for processing messages
        context_injector_params (PipesParams): Parameters yielded by the context injector,
            indicating the location from which the external process should load context data.
        message_reader_params (PipesParams): Parameters yielded by the message reader, indicating
            the location to which the external process should write messages.
        created_at (datetime): The time at which the session was created. Useful as cutoff for
            reading logs.
    """

    context_data: PipesContextData
    message_handler: PipesMessageHandler
    context_injector_params: PipesParams
    message_reader_params: PipesParams
    context: Union[OpExecutionContext, AssetExecutionContext]
    created_at: datetime = field(default_factory=datetime.now)

    @cached_property
    def default_remote_invocation_info(self) -> dict[str, str]:
        return {**self.context.dagster_run.dagster_execution_info}

    @public
    def get_bootstrap_env_vars(self) -> Mapping[str, str]:
        """Encode context injector and message reader params as environment variables.

        Passing environment variables is the typical way to expose the pipes I/O parameters
        to a pipes process.

        Returns:
            Mapping[str, str]: Environment variables to pass to the external process. The values are
            serialized as json, compressed with gzip, and then base-64-encoded.
        """
        return {
            param_name: encode_param(param_value)
            for param_name, param_value in self.get_bootstrap_params().items()
        }

    @public
    def get_bootstrap_cli_arguments(self) -> Mapping[str, str]:
        """Encode context injector and message reader params as CLI arguments.

        Passing CLI arguments is an alternative way to expose the pipes I/O parameters to a pipes process.
        Using environment variables should be preferred when possible.

        Returns:
            Mapping[str, str]: CLI arguments pass to the external process. The values are
            serialized as json, compressed with zlib, and then base64-encoded.
        """
        return {
            _env_var_to_cli_argument(param_name): encode_param(param_value)
            for param_name, param_value in self.get_bootstrap_params().items()
        }

    @public
    def get_bootstrap_params(self) -> Mapping[str, Any]:
        """Get the params necessary to bootstrap a launched pipes process. These parameters are typically
        are as environment variable. See `get_bootstrap_env_vars`. It is the context injector's
        responsibility to decide how to pass these parameters to the external environment.

        Returns:
            Mapping[str, str]: Parameters to pass to the external process and their corresponding
            values that must be passed by the context injector.
        """
        return {
            DAGSTER_PIPES_CONTEXT_ENV_VAR: self.context_injector_params,
            DAGSTER_PIPES_MESSAGES_ENV_VAR: self.message_reader_params,
        }

    @public
    def get_results(
        self,
        *,
        implicit_materializations: bool = True,
        metadata: Optional[Mapping[str, MetadataValue]] = None,
    ) -> Sequence[PipesExecutionResult]:
        """:py:class:`PipesExecutionResult` objects reported from the external process,
            potentially modified by Pipes.

        Args:
            implicit_materializations (bool): Create MaterializeResults for expected assets
                even was nothing is reported from the external process.
            metadata (Optional[Mapping[str, MetadataValue]]): Arbitrary metadata that will be attached to all
                results generated by the invocation. Useful for attaching information to
                asset materializations and checks that is available via the external process launch API
                but not in the external process itself (e.g. a job_id param returned by the launch API call).

        Returns:
            Sequence[PipesExecutionResult]: Result reported by external process.
        """
        metadata = metadata or {}

        reported = self.message_handler.get_reported_results()
        reported = tuple(self._inject_metadata_into_results(reported, metadata))

        if not implicit_materializations:
            return reported

        reported_keys = set(
            result.asset_key for result in reported if isinstance(result, MaterializeResult)
        )
        implicit = [
            MaterializeResult(asset_key=key)
            for key in self.context.selected_asset_keys
            if key not in reported_keys
        ]

        implicit = self._inject_metadata_into_results(implicit, metadata)

        return (*reported, *implicit)

    @public
    def get_reported_results(self) -> Sequence[PipesExecutionResult]:
        """:py:class:`PipesExecutionResult` objects only explicitly received from the external process.

        Returns:
            Sequence[PipesExecutionResult]: Result reported by external process.
        """
        return self.message_handler.get_reported_results()

    @public
    def get_custom_messages(self) -> Sequence[Any]:
        """Get the sequence of deserialized JSON data that was reported from the external process using
        `report_custom_message`.

        Returns: Sequence[Any]
        """
        return self.message_handler.get_custom_messages()

    def report_launched(self, launched_payload: PipesLaunchedData):
        """Submit arbitrary information about the launched process to Pipes.

        This hook is not necessarily called in every pipes session. It is useful primarily when we wish to
        condition the behavior of Pipes on some parameter that is only available after
        external process launch (such as a run id in the external system). The code calling `open_pipes_session()`
        is responsible for calling `PipesSession.report_launched()`.

        Passes the `launched_payload` to the message reader's `on_launched` method.
        """
        self.message_handler.on_launched(launched_payload)

    def _inject_metadata_into_results(
        self, results: Sequence[PipesExecutionResult], metadata: Mapping[str, MetadataValue]
    ) -> Sequence[PipesExecutionResult]:
        results_with_metadata = []
        for result in results:
            result = result._replace(metadata={**(result.metadata or {}), **metadata})  # noqa: PLW2901
            results_with_metadata.append(result)

        return results_with_metadata


def build_external_execution_context_data(
    context: Union[OpExecutionContext, AssetExecutionContext],
    extras: Optional[PipesExtras],
) -> "PipesContextData":
    asset_keys = (
        [_convert_asset_key(key) for key in sorted(context.selected_asset_keys)]
        if context.has_assets_def
        else None
    )
    code_version_by_asset_key = (
        {
            _convert_asset_key(key): context.assets_def.code_versions_by_key[key]
            for key in context.selected_asset_keys
        }
        if context.has_assets_def
        else None
    )
    provenance_by_asset_key = (
        {
            _convert_asset_key(key): _convert_data_provenance(context.get_asset_provenance(key))
            for key in context.selected_asset_keys
        }
        if context.has_assets_def
        else None
    )
    partition_key = context.partition_key if context.has_partition_key else None
    partition_key_range = context.partition_key_range if context.has_partition_key else None
    partition_time_window = (
        context.partition_time_window
        if context.has_partition_key
        and has_one_dimension_time_window_partitioning(
            context.get_step_execution_context().run_partitions_def
        )
        else None
    )
    return PipesContextData(
        asset_keys=asset_keys,
        code_version_by_asset_key=code_version_by_asset_key,
        provenance_by_asset_key=provenance_by_asset_key,
        partition_key=partition_key,
        partition_key_range=(
            _convert_partition_key_range(partition_key_range) if partition_key_range else None
        ),
        partition_time_window=(
            _convert_time_window(partition_time_window) if partition_time_window else None
        ),
        run_id=context.run_id,
        job_name=None if isinstance(context, BaseDirectExecutionContext) else context.job_name,
        retry_number=0 if isinstance(context, BaseDirectExecutionContext) else context.retry_number,
        extras=extras or {},
    )


def _convert_asset_key(asset_key: AssetKey) -> str:
    r"""Convert asset key to Pipes-compatible string representation.

    This includes escaping forward slashes (/) with backslashes (\).
    """
    return asset_key.to_escaped_user_string()


def _convert_data_provenance(
    provenance: Optional[DataProvenance],
) -> Optional["PipesDataProvenance"]:
    return (
        None
        if provenance is None
        else PipesDataProvenance(
            code_version=provenance.code_version,
            input_data_versions={
                _convert_asset_key(k): v.value for k, v in provenance.input_data_versions.items()
            },
            is_user_provided=provenance.is_user_provided,
        )
    )


def _convert_time_window(
    time_window: TimeWindow,
) -> "PipesTimeWindow":
    return PipesTimeWindow(
        start=time_window.start.isoformat(),
        end=time_window.end.isoformat(),
    )


def _convert_partition_key_range(
    partition_key_range: PartitionKeyRange,
) -> "PipesTimeWindow":
    return PipesTimeWindow(
        start=partition_key_range.start,
        end=partition_key_range.end,
    )


def _ser_err_from_pipes_exc(exc: PipesException):
    return SerializableErrorInfo(
        message=exc["message"],
        stack=exc["stack"],
        cls_name=exc["name"],
        cause=_ser_err_from_pipes_exc(exc["cause"]) if exc["cause"] else None,
        context=_ser_err_from_pipes_exc(exc["context"]) if exc["context"] else None,
    )



from collections import defaultdict
from collections.abc import Iterable, Mapping, Sequence
from functools import cached_property
from typing import AbstractSet, Optional, Union  # noqa: UP035

from dagster._core.definitions.asset_check_spec import AssetCheckKey, AssetCheckSpec
from dagster._core.definitions.asset_spec import (
    SYSTEM_METADATA_KEY_AUTO_CREATED_STUB_ASSET,
    AssetExecutionType,
    AssetSpec,
)
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.auto_materialize_policy import AutoMaterializePolicy
from dagster._core.definitions.backfill_policy import BackfillPolicy
from dagster._core.definitions.base_asset_graph import (
    AssetCheckNode,
    BaseAssetGraph,
    BaseAssetNode,
    EntityKey,
)
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
)
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.freshness_policy import FreshnessPolicy
from dagster._core.definitions.metadata import ArbitraryMetadataMapping
from dagster._core.definitions.partition import PartitionsDefinition
from dagster._core.definitions.partition_mapping import PartitionMapping
from dagster._core.definitions.resolved_asset_deps import ResolvedAssetDependencies
from dagster._core.definitions.source_asset import SourceAsset
from dagster._core.definitions.utils import DEFAULT_GROUP_NAME
from dagster._core.selector.subset_selector import generate_asset_dep_graph
from dagster._utils.warnings import disable_dagster_warnings


class AssetNode(BaseAssetNode):
    def __init__(
        self,
        key: AssetKey,
        parent_keys: AbstractSet[AssetKey],
        child_keys: AbstractSet[AssetKey],
        assets_def: AssetsDefinition,
        check_keys: AbstractSet[AssetCheckKey],
    ):
        self.key = key
        self.parent_keys = parent_keys
        self.child_keys = child_keys
        self.assets_def = assets_def
        self._check_keys = check_keys
        self._spec = assets_def.specs_by_key[key]

    @property
    def description(self) -> Optional[str]:
        return self._spec.description

    @property
    def group_name(self) -> str:
        return self._spec.group_name or DEFAULT_GROUP_NAME

    @property
    def is_materializable(self) -> bool:
        return self.assets_def.is_materializable

    @property
    def is_observable(self) -> bool:
        return self.assets_def.is_observable

    @property
    def is_external(self) -> bool:
        return self.assets_def.is_external

    @property
    def is_executable(self) -> bool:
        return self.assets_def.is_executable

    @property
    def metadata(self) -> ArbitraryMetadataMapping:
        return self._spec.metadata

    @property
    def tags(self) -> Mapping[str, str]:
        return self._spec.tags

    @property
    def kinds(self) -> AbstractSet[str]:
        return self._spec.kinds or set()

    @property
    def pools(self) -> Optional[set[str]]:
        if not self.assets_def.computation:
            return None
        return set(
            op_def.pool
            for op_def in self.assets_def.computation.node_def.iterate_op_defs()
            if op_def.pool
        )

    @property
    def owners(self) -> Sequence[str]:
        return self._spec.owners

    @property
    def is_partitioned(self) -> bool:
        return self.partitions_def is not None

    @property
    def partitions_def(self) -> Optional[PartitionsDefinition]:
        return self.assets_def.specs_by_key[self.key].partitions_def

    @property
    def partition_mappings(self) -> Mapping[AssetKey, PartitionMapping]:
        return self._spec.partition_mappings

    @property
    def freshness_policy(self) -> Optional[FreshnessPolicy]:
        return self._spec.freshness_policy

    @property
    def auto_materialize_policy(self) -> Optional[AutoMaterializePolicy]:
        return self._spec.auto_materialize_policy

    @property
    def automation_condition(self) -> Optional[AutomationCondition]:
        return self._spec.automation_condition

    @property
    def auto_observe_interval_minutes(self) -> Optional[float]:
        return self.assets_def.auto_observe_interval_minutes

    @property
    def backfill_policy(self) -> Optional[BackfillPolicy]:
        return self.assets_def.backfill_policy

    @property
    def code_version(self) -> Optional[str]:
        return self._spec.code_version

    @property
    def check_keys(self) -> AbstractSet[AssetCheckKey]:
        return self._check_keys

    @property
    def execution_set_asset_keys(self) -> AbstractSet[AssetKey]:
        return (
            {self.key}
            if len(self.assets_def.keys) <= 1 or self.assets_def.can_subset
            else self.assets_def.keys
        )

    @property
    def execution_set_entity_keys(self) -> AbstractSet[EntityKey]:
        if self.assets_def.can_subset:
            return {self.key}
        else:
            return self.assets_def.asset_and_check_keys

    ##### ASSET GRAPH SPECIFIC INTERFACE

    @property
    def execution_type(self) -> AssetExecutionType:
        return self.assets_def.execution_type

    @property
    def io_manager_key(self) -> str:
        return self.assets_def.get_io_manager_key_for_asset_key(self.key)

    def to_asset_spec(self) -> AssetSpec:
        return self._spec


class AssetGraph(BaseAssetGraph[AssetNode]):
    _assets_defs_by_check_key: Mapping[AssetCheckKey, AssetsDefinition]

    def __init__(
        self,
        asset_nodes_by_key: Mapping[AssetKey, AssetNode],
        assets_defs_by_check_key: Mapping[AssetCheckKey, AssetsDefinition],
    ):
        self._asset_nodes_by_key = asset_nodes_by_key
        self._asset_check_nodes_by_key = {
            k: AssetCheckNode(
                k,
                [d.asset_key for d in v.get_spec_for_check_key(k).additional_deps],
                v.get_spec_for_check_key(k).blocking,
                v.get_spec_for_check_key(k).automation_condition,
            )
            for k, v in assets_defs_by_check_key.items()
        }
        self._assets_defs_by_check_key = assets_defs_by_check_key

    @staticmethod
    def normalize_assets(
        assets: Iterable[Union[AssetsDefinition, SourceAsset]],
    ) -> Sequence[AssetsDefinition]:
        """Normalize a mixed list of AssetsDefinition and SourceAsset to a list of AssetsDefinition.

        Normalization includse:

        - Converting any SourceAsset to an AssetDefinition
        - Resolving all relative asset keys (that sometimes specify dependencies) to absolute asset
          keys
        - Creating unexecutable external asset definitions for any keys referenced by asset checks
          or as dependencies, but for which no definition was provided.
        """
        from dagster._core.definitions.external_asset import create_external_asset_from_source_asset

        # Convert any source assets to external assets
        assets_defs = [
            create_external_asset_from_source_asset(a) if isinstance(a, SourceAsset) else a
            for a in assets
        ]
        all_keys = {k for asset_def in assets_defs for k in asset_def.keys}

        # Resolve all asset dependencies. An asset dependency is resolved when its key is an
        # AssetKey not subject to any further manipulation.
        resolved_deps = ResolvedAssetDependencies(assets_defs, [])

        asset_key_replacements = [
            {
                raw_key: normalized_key
                for input_name, raw_key in ad.keys_by_input_name.items()
                if (
                    normalized_key := resolved_deps.get_resolved_asset_key_for_input(ad, input_name)
                )
                != raw_key
            }
            for ad in assets_defs
        ]

        # Only update the assets defs if we're actually replacing input asset keys
        assets_defs = [
            ad.with_attributes(asset_key_replacements=reps) if reps else ad
            for ad, reps in zip(assets_defs, asset_key_replacements)
        ]

        # Create unexecutable external assets definitions for any referenced keys for which no
        # definition was provided.
        all_referenced_asset_keys = {
            key for assets_def in assets_defs for key in assets_def.dependency_keys
        }
        with disable_dagster_warnings():
            for key in all_referenced_asset_keys.difference(all_keys):
                assets_defs.append(
                    AssetsDefinition(
                        specs=[
                            AssetSpec(
                                key=key,
                                metadata={SYSTEM_METADATA_KEY_AUTO_CREATED_STUB_ASSET: True},
                            )
                        ]
                    )
                )
        return assets_defs

    @classmethod
    def from_assets(
        cls,
        assets: Iterable[Union[AssetsDefinition, SourceAsset]],
    ) -> "AssetGraph":
        assets_defs = cls.normalize_assets(assets)

        # Build the set of AssetNodes. Each node holds key rather than object references to parent
        # and child nodes.
        dep_graph = generate_asset_dep_graph(assets_defs)

        assets_defs_by_check_key: dict[AssetCheckKey, AssetsDefinition] = {}
        check_keys_by_asset_key: defaultdict[AssetKey, set[AssetCheckKey]] = defaultdict(set)
        for ad in assets_defs:
            for ck in ad.check_keys:
                check_keys_by_asset_key[ck.asset_key].add(ck)
                assets_defs_by_check_key[ck] = ad

        asset_nodes_by_key = {
            key: AssetNode(
                key=key,
                parent_keys=dep_graph["upstream"][key],
                child_keys=dep_graph["downstream"][key],
                assets_def=ad,
                check_keys=check_keys_by_asset_key[key],
            )
            for ad in assets_defs
            for key in ad.keys
        }

        return AssetGraph(
            asset_nodes_by_key=asset_nodes_by_key,
            assets_defs_by_check_key=assets_defs_by_check_key,
        )

    def get_execution_set_asset_and_check_keys(
        self, entity_key: EntityKey
    ) -> AbstractSet[EntityKey]:
        if isinstance(entity_key, AssetKey):
            return self.get(entity_key).execution_set_entity_keys
        else:  # AssetCheckKey
            assets_def = self._assets_defs_by_check_key[entity_key]
            return {entity_key} if assets_def.can_subset else assets_def.asset_and_check_keys

    def get_execution_set_identifier(self, entity_key: EntityKey) -> Optional[str]:
        """All assets and asset checks with the same execution_set_identifier must be executed
        together - i.e. you can't execute just a subset of them.

        Returns None if the asset or asset check can be executed individually.
        """
        from dagster._core.definitions.graph_definition import GraphDefinition

        if isinstance(entity_key, AssetKey):
            assets_def = self.get(entity_key).assets_def
            return (
                assets_def.unique_id
                if (
                    assets_def.is_executable
                    and not assets_def.can_subset
                    and (len(assets_def.keys) > 1 or assets_def.check_keys)
                )
                else None
            )

        else:  # AssetCheckKey
            assets_def = self._assets_defs_by_check_key[entity_key]
            # Executing individual checks isn't supported in graph assets
            if isinstance(assets_def.node_def, GraphDefinition):
                return assets_def.unique_id
            else:
                return assets_def.unique_id if not assets_def.can_subset else None

    @cached_property
    def assets_defs(self) -> Sequence[AssetsDefinition]:
        return list(
            {
                *(asset.assets_def for asset in self.asset_nodes),
                *(ad for ad in self._assets_defs_by_check_key.values()),
            }
        )

    def assets_defs_for_keys(self, keys: Iterable[EntityKey]) -> Sequence[AssetsDefinition]:
        return list({self.assets_def_for_key(key) for key in keys})

    def assets_def_for_key(self, key: EntityKey) -> AssetsDefinition:
        if isinstance(key, AssetKey):
            return self.get(key).assets_def
        else:
            return self._assets_defs_by_check_key[key]

    @cached_property
    def asset_check_keys(self) -> AbstractSet[AssetCheckKey]:
        return {key for ad in self.assets_defs for key in ad.check_keys}

    @cached_property
    def unpartitioned_assets_def_asset_check_keys(self) -> AbstractSet[AssetCheckKey]:
        """Asset check keys that correspond to unpartitioned AssetsDefinitions."""
        return {
            k for k in self.asset_check_keys if self.assets_def_for_key(k).partitions_def is None
        }

    def get_check_spec(self, key: AssetCheckKey) -> AssetCheckSpec:
        return self._assets_defs_by_check_key[key].get_spec_for_check_key(key)


def executable_in_same_run(
    asset_graph: BaseAssetGraph, child_key: EntityKey, parent_key: EntityKey
):
    """Returns whether a child asset can be materialized in the same run as a parent asset."""
    from dagster._core.definitions.partition_mapping import IdentityPartitionMapping
    from dagster._core.definitions.remote_asset_graph import RemoteWorkspaceAssetGraph
    from dagster._core.definitions.time_window_partition_mapping import TimeWindowPartitionMapping

    child_node = asset_graph.get(child_key)
    parent_node = asset_graph.get(parent_key)

    # if operating on a RemoteWorkspaceAssetGraph, must share a repository handle
    if isinstance(asset_graph, RemoteWorkspaceAssetGraph):
        child_handle = asset_graph.get_repository_handle(child_key)
        parent_handle = asset_graph.get_repository_handle(parent_key)
        if child_handle != parent_handle:
            return False

    # partitions definitions must match
    if child_node.partitions_def != parent_node.partitions_def:
        return False

    # unpartitioned assets can always execute together
    if child_node.partitions_def is None:
        return True

    return isinstance(
        asset_graph.get_partition_mapping(child_node.key, parent_node.key),
        (TimeWindowPartitionMapping, IdentityPartitionMapping),
    )



# flake8: noqa
# type: ignore
# Generated from AssetSelection.g4 by ANTLR 4.13.2
from antlr4 import *
from .AssetSelectionParser import AssetSelectionParser


# This class defines a complete listener for a parse tree produced by AssetSelectionParser.
class AssetSelectionListener(ParseTreeListener):
    # Enter a parse tree produced by AssetSelectionParser#start.
    def enterStart(self, ctx: AssetSelectionParser.StartContext):
        pass

    # Exit a parse tree produced by AssetSelectionParser#start.
    def exitStart(self, ctx: AssetSelectionParser.StartContext):
        pass

    # Enter a parse tree produced by AssetSelectionParser#UpTraversalExpression.
    def enterUpTraversalExpression(self, ctx: AssetSelectionParser.UpTraversalExpressionContext):
        pass

    # Exit a parse tree produced by AssetSelectionParser#UpTraversalExpression.
    def exitUpTraversalExpression(self, ctx: AssetSelectionParser.UpTraversalExpressionContext):
        pass

    # Enter a parse tree produced by AssetSelectionParser#AndExpression.
    def enterAndExpression(self, ctx: AssetSelectionParser.AndExpressionContext):
        pass

    # Exit a parse tree produced by AssetSelectionParser#AndExpression.
    def exitAndExpression(self, ctx: AssetSelectionParser.AndExpressionContext):
        pass

    # Enter a parse tree produced by AssetSelectionParser#AllExpression.
    def enterAllExpression(self, ctx: AssetSelectionParser.AllExpressionContext):
        pass

    # Exit a parse tree produced by AssetSelectionParser#AllExpression.
    def exitAllExpression(self, ctx: AssetSelectionParser.AllExpressionContext):
        pass

    # Enter a parse tree produced by AssetSelectionParser#TraversalAllowedExpression.
    def enterTraversalAllowedExpression(
        self, ctx: AssetSelectionParser.TraversalAllowedExpressionContext
    ):
        pass

    # Exit a parse tree produced by AssetSelectionParser#TraversalAllowedExpression.
    def exitTraversalAllowedExpression(
        self, ctx: AssetSelectionParser.TraversalAllowedExpressionContext
    ):
        pass

    # Enter a parse tree produced by AssetSelectionParser#DownTraversalExpression.
    def enterDownTraversalExpression(
        self, ctx: AssetSelectionParser.DownTraversalExpressionContext
    ):
        pass

    # Exit a parse tree produced by AssetSelectionParser#DownTraversalExpression.
    def exitDownTraversalExpression(self, ctx: AssetSelectionParser.DownTraversalExpressionContext):
        pass

    # Enter a parse tree produced by AssetSelectionParser#NotExpression.
    def enterNotExpression(self, ctx: AssetSelectionParser.NotExpressionContext):
        pass

    # Exit a parse tree produced by AssetSelectionParser#NotExpression.
    def exitNotExpression(self, ctx: AssetSelectionParser.NotExpressionContext):
        pass

    # Enter a parse tree produced by AssetSelectionParser#OrExpression.
    def enterOrExpression(self, ctx: AssetSelectionParser.OrExpressionContext):
        pass

    # Exit a parse tree produced by AssetSelectionParser#OrExpression.
    def exitOrExpression(self, ctx: AssetSelectionParser.OrExpressionContext):
        pass

    # Enter a parse tree produced by AssetSelectionParser#UpAndDownTraversalExpression.
    def enterUpAndDownTraversalExpression(
        self, ctx: AssetSelectionParser.UpAndDownTraversalExpressionContext
    ):
        pass

    # Exit a parse tree produced by AssetSelectionParser#UpAndDownTraversalExpression.
    def exitUpAndDownTraversalExpression(
        self, ctx: AssetSelectionParser.UpAndDownTraversalExpressionContext
    ):
        pass

    # Enter a parse tree produced by AssetSelectionParser#AttributeExpression.
    def enterAttributeExpression(self, ctx: AssetSelectionParser.AttributeExpressionContext):
        pass

    # Exit a parse tree produced by AssetSelectionParser#AttributeExpression.
    def exitAttributeExpression(self, ctx: AssetSelectionParser.AttributeExpressionContext):
        pass

    # Enter a parse tree produced by AssetSelectionParser#FunctionCallExpression.
    def enterFunctionCallExpression(self, ctx: AssetSelectionParser.FunctionCallExpressionContext):
        pass

    # Exit a parse tree produced by AssetSelectionParser#FunctionCallExpression.
    def exitFunctionCallExpression(self, ctx: AssetSelectionParser.FunctionCallExpressionContext):
        pass

    # Enter a parse tree produced by AssetSelectionParser#ParenthesizedExpression.
    def enterParenthesizedExpression(
        self, ctx: AssetSelectionParser.ParenthesizedExpressionContext
    ):
        pass

    # Exit a parse tree produced by AssetSelectionParser#ParenthesizedExpression.
    def exitParenthesizedExpression(self, ctx: AssetSelectionParser.ParenthesizedExpressionContext):
        pass

    # Enter a parse tree produced by AssetSelectionParser#upTraversal.
    def enterUpTraversal(self, ctx: AssetSelectionParser.UpTraversalContext):
        pass

    # Exit a parse tree produced by AssetSelectionParser#upTraversal.
    def exitUpTraversal(self, ctx: AssetSelectionParser.UpTraversalContext):
        pass

    # Enter a parse tree produced by AssetSelectionParser#downTraversal.
    def enterDownTraversal(self, ctx: AssetSelectionParser.DownTraversalContext):
        pass

    # Exit a parse tree produced by AssetSelectionParser#downTraversal.
    def exitDownTraversal(self, ctx: AssetSelectionParser.DownTraversalContext):
        pass

    # Enter a parse tree produced by AssetSelectionParser#functionName.
    def enterFunctionName(self, ctx: AssetSelectionParser.FunctionNameContext):
        pass

    # Exit a parse tree produced by AssetSelectionParser#functionName.
    def exitFunctionName(self, ctx: AssetSelectionParser.FunctionNameContext):
        pass

    # Enter a parse tree produced by AssetSelectionParser#KeyExpr.
    def enterKeyExpr(self, ctx: AssetSelectionParser.KeyExprContext):
        pass

    # Exit a parse tree produced by AssetSelectionParser#KeyExpr.
    def exitKeyExpr(self, ctx: AssetSelectionParser.KeyExprContext):
        pass

    # Enter a parse tree produced by AssetSelectionParser#TagAttributeExpr.
    def enterTagAttributeExpr(self, ctx: AssetSelectionParser.TagAttributeExprContext):
        pass

    # Exit a parse tree produced by AssetSelectionParser#TagAttributeExpr.
    def exitTagAttributeExpr(self, ctx: AssetSelectionParser.TagAttributeExprContext):
        pass

    # Enter a parse tree produced by AssetSelectionParser#OwnerAttributeExpr.
    def enterOwnerAttributeExpr(self, ctx: AssetSelectionParser.OwnerAttributeExprContext):
        pass

    # Exit a parse tree produced by AssetSelectionParser#OwnerAttributeExpr.
    def exitOwnerAttributeExpr(self, ctx: AssetSelectionParser.OwnerAttributeExprContext):
        pass

    # Enter a parse tree produced by AssetSelectionParser#GroupAttributeExpr.
    def enterGroupAttributeExpr(self, ctx: AssetSelectionParser.GroupAttributeExprContext):
        pass

    # Exit a parse tree produced by AssetSelectionParser#GroupAttributeExpr.
    def exitGroupAttributeExpr(self, ctx: AssetSelectionParser.GroupAttributeExprContext):
        pass

    # Enter a parse tree produced by AssetSelectionParser#KindAttributeExpr.
    def enterKindAttributeExpr(self, ctx: AssetSelectionParser.KindAttributeExprContext):
        pass

    # Exit a parse tree produced by AssetSelectionParser#KindAttributeExpr.
    def exitKindAttributeExpr(self, ctx: AssetSelectionParser.KindAttributeExprContext):
        pass

    # Enter a parse tree produced by AssetSelectionParser#ColumnAttributeExpr.
    def enterColumnAttributeExpr(self, ctx: AssetSelectionParser.ColumnAttributeExprContext):
        pass

    # Exit a parse tree produced by AssetSelectionParser#ColumnAttributeExpr.
    def exitColumnAttributeExpr(self, ctx: AssetSelectionParser.ColumnAttributeExprContext):
        pass

    # Enter a parse tree produced by AssetSelectionParser#TableNameAttributeExpr.
    def enterTableNameAttributeExpr(self, ctx: AssetSelectionParser.TableNameAttributeExprContext):
        pass

    # Exit a parse tree produced by AssetSelectionParser#TableNameAttributeExpr.
    def exitTableNameAttributeExpr(self, ctx: AssetSelectionParser.TableNameAttributeExprContext):
        pass

    # Enter a parse tree produced by AssetSelectionParser#ColumnTagAttributeExpr.
    def enterColumnTagAttributeExpr(self, ctx: AssetSelectionParser.ColumnTagAttributeExprContext):
        pass

    # Exit a parse tree produced by AssetSelectionParser#ColumnTagAttributeExpr.
    def exitColumnTagAttributeExpr(self, ctx: AssetSelectionParser.ColumnTagAttributeExprContext):
        pass

    # Enter a parse tree produced by AssetSelectionParser#CodeLocationAttributeExpr.
    def enterCodeLocationAttributeExpr(
        self, ctx: AssetSelectionParser.CodeLocationAttributeExprContext
    ):
        pass

    # Exit a parse tree produced by AssetSelectionParser#CodeLocationAttributeExpr.
    def exitCodeLocationAttributeExpr(
        self, ctx: AssetSelectionParser.CodeLocationAttributeExprContext
    ):
        pass

    # Enter a parse tree produced by AssetSelectionParser#ChangedInBranchAttributeExpr.
    def enterChangedInBranchAttributeExpr(
        self, ctx: AssetSelectionParser.ChangedInBranchAttributeExprContext
    ):
        pass

    # Exit a parse tree produced by AssetSelectionParser#ChangedInBranchAttributeExpr.
    def exitChangedInBranchAttributeExpr(
        self, ctx: AssetSelectionParser.ChangedInBranchAttributeExprContext
    ):
        pass

    # Enter a parse tree produced by AssetSelectionParser#value.
    def enterValue(self, ctx: AssetSelectionParser.ValueContext):
        pass

    # Exit a parse tree produced by AssetSelectionParser#value.
    def exitValue(self, ctx: AssetSelectionParser.ValueContext):
        pass

    # Enter a parse tree produced by AssetSelectionParser#keyValue.
    def enterKeyValue(self, ctx: AssetSelectionParser.KeyValueContext):
        pass

    # Exit a parse tree produced by AssetSelectionParser#keyValue.
    def exitKeyValue(self, ctx: AssetSelectionParser.KeyValueContext):
        pass


del AssetSelectionParser



from dagster._core.launcher.base import (
    CheckRunHealthResult as CheckRunHealthResult,
    LaunchRunContext as LaunchRunContext,
    ResumeRunContext as ResumeRunContext,
    RunLauncher as RunLauncher,
    WorkerStatus as WorkerStatus,
)
from dagster._core.launcher.default_run_launcher import DefaultRunLauncher as DefaultRunLauncher



from antlr4 import CommonTokenStream, InputStream
from antlr4.error.ErrorListener import ErrorListener

from dagster._core.definitions.antlr_asset_selection.generated.AssetSelectionLexer import (
    AssetSelectionLexer,
)
from dagster._core.definitions.antlr_asset_selection.generated.AssetSelectionParser import (
    AssetSelectionParser,
)
from dagster._core.definitions.antlr_asset_selection.generated.AssetSelectionVisitor import (
    AssetSelectionVisitor,
)
from dagster._core.definitions.asset_selection import (
    AssetSelection,
    ChangedInBranchAssetSelection,
    CodeLocationAssetSelection,
    ColumnAssetSelection,
    ColumnTagAssetSelection,
    KeyWildCardAssetSelection,
    TableNameAssetSelection,
)
from dagster._core.storage.tags import KIND_PREFIX


class AntlrInputErrorListener(ErrorListener):
    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        raise Exception(f"Syntax error at line {line}, column {column}: {msg}")


def parse_traversal_depth(optional_digits):
    """Return an integer from DIGITS text if present, otherwise None (meaning infinite depth)."""
    if optional_digits is None:
        return None
    return int(optional_digits.getText())


class AntlrAssetSelectionVisitor(AssetSelectionVisitor):
    def __init__(self, include_sources: bool):
        self.include_sources = include_sources

    def visitStart(self, ctx: AssetSelectionParser.StartContext):
        return self.visit(ctx.expr())

    def visitTraversalAllowedExpression(
        self, ctx: AssetSelectionParser.TraversalAllowedExpressionContext
    ):
        return self.visit(ctx.traversalAllowedExpr())

    def visitUpAndDownTraversalExpression(
        self, ctx: AssetSelectionParser.UpAndDownTraversalExpressionContext
    ):
        selection: AssetSelection = self.visit(ctx.traversalAllowedExpr())
        # upTraversal => optional DIGITS? PLUS
        up_digits = ctx.upTraversal().DIGITS()
        up_depth = parse_traversal_depth(up_digits)

        # downTraversal => PLUS DIGITS?
        down_digits = ctx.downTraversal().DIGITS()
        down_depth = parse_traversal_depth(down_digits)

        return selection.upstream(depth=up_depth) | selection.downstream(depth=down_depth)

    def visitUpTraversalExpression(self, ctx: AssetSelectionParser.UpTraversalExpressionContext):
        selection: AssetSelection = self.visit(ctx.traversalAllowedExpr())
        up_digits = ctx.upTraversal().DIGITS()
        up_depth = parse_traversal_depth(up_digits)
        return selection.upstream(depth=up_depth)

    def visitDownTraversalExpression(
        self, ctx: AssetSelectionParser.DownTraversalExpressionContext
    ):
        selection: AssetSelection = self.visit(ctx.traversalAllowedExpr())
        down_digits = ctx.downTraversal().DIGITS()
        down_depth = parse_traversal_depth(down_digits)
        return selection.downstream(depth=down_depth)

    def visitNotExpression(self, ctx: AssetSelectionParser.NotExpressionContext):
        selection: AssetSelection = self.visit(ctx.expr())
        return AssetSelection.all(include_sources=self.include_sources) - selection

    def visitAndExpression(self, ctx: AssetSelectionParser.AndExpressionContext):
        left: AssetSelection = self.visit(ctx.expr(0))
        right: AssetSelection = self.visit(ctx.expr(1))
        return left & right

    def visitOrExpression(self, ctx: AssetSelectionParser.OrExpressionContext):
        left: AssetSelection = self.visit(ctx.expr(0))
        right: AssetSelection = self.visit(ctx.expr(1))
        return left | right

    def visitAllExpression(self, ctx: AssetSelectionParser.AllExpressionContext):
        return AssetSelection.all(include_sources=self.include_sources)

    def visitAttributeExpression(self, ctx: AssetSelectionParser.AttributeExpressionContext):
        return self.visit(ctx.attributeExpr())

    def visitFunctionCallExpression(self, ctx: AssetSelectionParser.FunctionCallExpressionContext):
        function = self.visit(ctx.functionName())
        selection: AssetSelection = self.visit(ctx.expr())
        if function == "sinks":
            return selection.sinks()
        elif function == "roots":
            return selection.roots()

    def visitParenthesizedExpression(
        self, ctx: AssetSelectionParser.ParenthesizedExpressionContext
    ):
        return self.visit(ctx.expr())

    def visitFunctionName(self, ctx: AssetSelectionParser.FunctionNameContext):
        if ctx.SINKS():
            return "sinks"
        elif ctx.ROOTS():
            return "roots"

    def visitKeyExpr(self, ctx: AssetSelectionParser.KeyExprContext):
        value = self.visit(ctx.keyValue())
        return KeyWildCardAssetSelection(selected_key_wildcard=value)

    def visitTagAttributeExpr(self, ctx: AssetSelectionParser.TagAttributeExprContext):
        key = self.visit(ctx.value(0))
        value = self.visit(ctx.value(1)) if ctx.EQUAL() else ""
        return AssetSelection.tag(key, value, include_sources=self.include_sources)

    def visitOwnerAttributeExpr(self, ctx: AssetSelectionParser.OwnerAttributeExprContext):
        owner = self.visit(ctx.value())
        return AssetSelection.owner(owner)

    def visitGroupAttributeExpr(self, ctx: AssetSelectionParser.GroupAttributeExprContext):
        group = self.visit(ctx.value())
        return AssetSelection.groups(group, include_sources=self.include_sources)

    def visitKindAttributeExpr(self, ctx: AssetSelectionParser.KindAttributeExprContext):
        kind = self.visit(ctx.value())
        return AssetSelection.tag(f"{KIND_PREFIX}{kind}", "", include_sources=self.include_sources)

    def visitCodeLocationAttributeExpr(
        self, ctx: AssetSelectionParser.CodeLocationAttributeExprContext
    ):
        code_location = self.visit(ctx.value())
        return CodeLocationAssetSelection(selected_code_location=code_location)

    def visitKeyValue(self, ctx: AssetSelectionParser.KeyValueContext):
        if ctx.QUOTED_STRING():
            return ctx.QUOTED_STRING().getText().strip('"')
        elif ctx.UNQUOTED_WILDCARD_STRING():
            return ctx.UNQUOTED_WILDCARD_STRING().getText()
        elif ctx.UNQUOTED_STRING():
            return ctx.UNQUOTED_STRING().getText()

    def visitValue(self, ctx: AssetSelectionParser.ValueContext):
        if ctx.QUOTED_STRING():
            return ctx.QUOTED_STRING().getText().strip('"')
        elif ctx.UNQUOTED_STRING():
            return ctx.UNQUOTED_STRING().getText()

    def visitColumnAttributeExpr(self, ctx: AssetSelectionParser.ColumnAttributeExprContext):
        column = self.visit(ctx.value())
        return ColumnAssetSelection(selected_column=column)

    def visitTableNameAttributeExpr(self, ctx: AssetSelectionParser.TableNameAttributeExprContext):
        table_name = self.visit(ctx.value())
        return TableNameAssetSelection(selected_table_name=table_name)

    def visitColumnTagAttributeExpr(self, ctx: AssetSelectionParser.ColumnTagAttributeExprContext):
        key = self.visit(ctx.value(0))
        value = self.visit(ctx.value(1)) if ctx.EQUAL() else ""
        return ColumnTagAssetSelection(key=key, value=value)

    def visitChangedInBranchAttributeExpr(
        self, ctx: AssetSelectionParser.ChangedInBranchAttributeExprContext
    ):
        branch = self.visit(ctx.value())
        return ChangedInBranchAssetSelection(selected_changed_in_branch=branch)


class AntlrAssetSelectionParser:
    def __init__(self, selection_str: str, include_sources: bool = False):
        lexer = AssetSelectionLexer(InputStream(selection_str))
        lexer.removeErrorListeners()  # Remove the default listener that just writes to the console
        lexer.addErrorListener(AntlrInputErrorListener())

        stream = CommonTokenStream(lexer)

        parser = AssetSelectionParser(stream)
        parser.removeErrorListeners()  # Remove the default listener that just writes to the console
        parser.addErrorListener(AntlrInputErrorListener())

        self._tree = parser.start()
        self._tree_str = self._tree.toStringTree(recog=parser)
        self._asset_selection = AntlrAssetSelectionVisitor(include_sources).visit(self._tree)

    @property
    def tree_str(self) -> str:
        return self._tree_str

    @property
    def asset_selection(self) -> AssetSelection:
        return self._asset_selection



from dataclasses import dataclass, replace
from typing import Generic, Optional, Union

import dagster._check as check
from dagster._core.definitions.asset_key import T_EntityKey
from dagster._core.definitions.events import AssetKeyPartitionKey
from dagster._core.definitions.partition import (
    AllPartitionsSubset,
    PartitionsDefinition,
    PartitionsSubset,
)
from dagster._core.definitions.time_window_partitions import TimeWindowPartitionsSubset
from dagster._serdes.serdes import DataclassSerializer, whitelist_for_serdes

EntitySubsetValue = Union[bool, PartitionsSubset]


class EntitySubsetSerializer(DataclassSerializer):
    """Ensures that the inner PartitionsSubset is converted to a serializable form if necessary."""

    def get_storage_name(self) -> str:
        # backcompat
        return "AssetSubset"

    def before_pack(self, value: "SerializableEntitySubset") -> "SerializableEntitySubset":
        if value.is_partitioned:
            return replace(value, value=value.subset_value.to_serializable_subset())
        return value


@whitelist_for_serdes(
    serializer=EntitySubsetSerializer,
    storage_field_names={"key": "asset_key"},
    old_storage_names={"AssetSubset"},
)
@dataclass(frozen=True)
class SerializableEntitySubset(Generic[T_EntityKey]):
    """Represents a serializable subset of a given EntityKey."""

    key: T_EntityKey
    value: EntitySubsetValue

    @property
    def is_partitioned(self) -> bool:
        return not isinstance(self.value, bool)

    @property
    def bool_value(self) -> bool:
        return check.inst(self.value, bool)

    @property
    def subset_value(self) -> PartitionsSubset:
        return check.inst(self.value, PartitionsSubset)

    @property
    def size(self) -> int:
        if not self.is_partitioned:
            return int(self.bool_value)
        else:
            return len(self.subset_value)

    @property
    def is_empty(self) -> bool:
        if self.is_partitioned:
            return self.subset_value.is_empty
        else:
            return not self.bool_value

    def is_compatible_with_partitions_def(
        self, partitions_def: Optional[PartitionsDefinition]
    ) -> bool:
        if self.is_partitioned:
            # for some PartitionSubset types, we have access to the underlying partitions
            # definitions, so we can ensure those are identical
            if isinstance(self.value, (TimeWindowPartitionsSubset, AllPartitionsSubset)):
                return self.value.partitions_def == partitions_def
            else:
                return partitions_def is not None
        else:
            return partitions_def is None

    def __contains__(self, item: AssetKeyPartitionKey) -> bool:
        if not self.is_partitioned:
            return item.asset_key == self.key and item.partition_key is None and self.bool_value
        else:
            return item.asset_key == self.key and item.partition_key in self.subset_value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}<{self.key}>({self.value})"



from collections.abc import Sequence
from typing import AbstractSet, Any, Optional  # noqa: UP035

import dagster._check as check
from dagster._annotations import public
from dagster._core.definitions.asset_check_spec import AssetCheckKey, AssetCheckSpec
from dagster._core.definitions.job_definition import JobDefinition
from dagster._core.definitions.op_definition import OpDefinition
from dagster._core.definitions.repository_definition.repository_definition import (
    RepositoryDefinition,
)
from dagster._core.definitions.step_launcher import StepLauncher
from dagster._core.errors import DagsterInvariantViolationError
from dagster._core.execution.context.asset_execution_context import (
    _copy_docs_from_op_execution_context,
)
from dagster._core.execution.context.op_execution_context import OpExecutionContext
from dagster._core.execution.context.system import StepExecutionContext
from dagster._core.instance import DagsterInstance
from dagster._core.log_manager import DagsterLogManager
from dagster._core.storage.dagster_run import DagsterRun
from dagster._utils.forked_pdb import ForkedPdb


class AssetCheckExecutionContext:
    def __init__(self, op_execution_context: OpExecutionContext) -> None:
        self._op_execution_context = check.inst_param(
            op_execution_context, "op_execution_context", OpExecutionContext
        )
        self._step_execution_context = self._op_execution_context._step_execution_context  # noqa: SLF001

    @staticmethod
    def get() -> "AssetCheckExecutionContext":
        from dagster._core.execution.context.compute import current_execution_context

        ctx = current_execution_context.get()
        if ctx is None:
            raise DagsterInvariantViolationError("No current AssetExecutionContext in scope.")
        if not isinstance(ctx, AssetCheckExecutionContext):
            raise DagsterInvariantViolationError(
                "Can't use AssetCheckExecutionContext.get() in the context of an "
                "AssetCheckExecutionContext or OpExecutionContext. Use AssetCheckExecutionContext.get() instead."
            )
        return ctx

    @property
    def op_execution_context(self) -> OpExecutionContext:
        return self._op_execution_context

    @public
    @property
    def log(self) -> DagsterLogManager:
        """The log manager available in the execution context. Logs will be viewable in the Dagster UI.
        Returns: DagsterLogManager.
        """
        return self.op_execution_context.log

    @public
    @property
    def pdb(self) -> ForkedPdb:
        """Gives access to pdb debugging from within the asset. Materializing the asset via the
        Dagster UI or CLI will enter the pdb debugging context in the process used to launch the UI or
        run the CLI.

        Returns: dagster.utils.forked_pdb.ForkedPdb
        """
        return self.op_execution_context.pdb

    @property
    def run(self) -> DagsterRun:
        """The DagsterRun object corresponding to the execution. Information like run_id,
        run configuration, and the assets selected for materialization can be found on the DagsterRun.
        """
        return self.op_execution_context.run

    @public
    @property
    def job_def(self) -> JobDefinition:
        """The definition for the currently executing job. Information like the job name, and job tags
        can be found on the JobDefinition.
        Returns: JobDefinition.
        """
        return self.op_execution_context.job_def

    @property
    def repository_def(self) -> RepositoryDefinition:
        """RepositoryDefinition: The Dagster repository for the currently executing job."""
        return self.op_execution_context.repository_def

    #### check related
    @public
    @property
    @_copy_docs_from_op_execution_context
    def selected_asset_check_keys(self) -> AbstractSet[AssetCheckKey]:
        return self.op_execution_context.selected_asset_check_keys

    @public
    @property
    def check_specs(self) -> Sequence[AssetCheckSpec]:
        """The asset check specs for the currently executing asset check."""
        return list(self.op_execution_context.assets_def.check_specs)

    #### op related

    @property
    @_copy_docs_from_op_execution_context
    def retry_number(self):
        return self.op_execution_context.retry_number

    @_copy_docs_from_op_execution_context
    def describe_op(self) -> str:
        return self.op_execution_context.describe_op()

    @public
    @property
    @_copy_docs_from_op_execution_context
    def op_def(self) -> OpDefinition:
        return self.op_execution_context.op_def

    #### execution related

    @public
    @property
    @_copy_docs_from_op_execution_context
    def instance(self) -> DagsterInstance:
        return self.op_execution_context.instance

    @property
    @_copy_docs_from_op_execution_context
    def step_launcher(self) -> Optional[StepLauncher]:
        return self.op_execution_context.step_launcher

    @_copy_docs_from_op_execution_context
    def get_step_execution_context(self) -> StepExecutionContext:
        return self.op_execution_context.get_step_execution_context()

    # misc

    @public
    @property
    @_copy_docs_from_op_execution_context
    def resources(self) -> Any:
        return self.op_execution_context.resources

    @property
    @_copy_docs_from_op_execution_context
    def is_subset(self):
        return self.op_execution_context.is_subset



from dagster._core.storage.event_log.sqlite.consolidated_sqlite_event_log import (
    ConsolidatedSqliteEventLogStorage as ConsolidatedSqliteEventLogStorage,
)
from dagster._core.storage.event_log.sqlite.sqlite_event_log import (
    SqliteEventLogStorage as SqliteEventLogStorage,
)



from typing import NamedTuple

from dagster._annotations import PublicAttr


class PartitionKeyRange(NamedTuple):
    """Defines a range of partitions.

    Args:
        start (str): The starting partition key in the range (inclusive).
        end (str): The ending partition key in the range (inclusive).

    Examples:
        .. code-block:: python

            partitions_def = StaticPartitionsDefinition(["a", "b", "c", "d"])
            partition_key_range = PartitionKeyRange(start="a", end="c") # Represents ["a", "b", "c"]
    """

    # Inclusive on both sides
    start: PublicAttr[str]
    end: PublicAttr[str]



import asyncio
import datetime
import logging
from collections import defaultdict
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, AbstractSet, Optional  # noqa: UP035

from dagster._core.asset_graph_view.asset_graph_view import AssetGraphView, TemporalContext
from dagster._core.asset_graph_view.entity_subset import EntitySubset
from dagster._core.definitions.asset_daemon_cursor import AssetDaemonCursor
from dagster._core.definitions.asset_key import EntityKey
from dagster._core.definitions.base_asset_graph import BaseAssetGraph, BaseAssetNode
from dagster._core.definitions.data_time import CachingDataTimeResolver
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
    AutomationResult,
)
from dagster._core.definitions.declarative_automation.automation_context import AutomationContext
from dagster._core.definitions.events import AssetKey
from dagster._core.instance import DagsterInstance
from dagster._time import get_current_datetime

if TYPE_CHECKING:
    from dagster._utils.caching_instance_queryer import CachingInstanceQueryer


class AutomationConditionEvaluator:
    def __init__(
        self,
        *,
        entity_keys: AbstractSet[EntityKey],
        instance: DagsterInstance,
        asset_graph: BaseAssetGraph,
        cursor: AssetDaemonCursor,
        emit_backfills: bool,
        default_condition: Optional[AutomationCondition] = None,
        evaluation_time: Optional[datetime.datetime] = None,
        logger: logging.Logger = logging.getLogger("dagster.automation"),
    ):
        self.entity_keys = entity_keys
        self.asset_graph_view = AssetGraphView(
            temporal_context=TemporalContext(
                effective_dt=evaluation_time or get_current_datetime(),
                last_event_id=instance.event_log_storage.get_maximum_record_id(),
            ),
            instance=instance,
            asset_graph=asset_graph,
        )
        self.logger = logger
        self.cursor = cursor
        self.default_condition = default_condition

        self.current_results_by_key: dict[EntityKey, AutomationResult] = {}
        self.condition_cursors = []
        self.expected_data_time_mapping = defaultdict()

        _instance = self.asset_graph_view.instance
        self.legacy_auto_materialize_run_tags: Mapping[str, str] = (
            _instance.auto_materialize_run_tags
        )
        self.legacy_respect_materialization_data_versions = (
            _instance.auto_materialize_respect_materialization_data_versions
        )
        self.emit_backfills = emit_backfills or _instance.da_request_backfills()

        self.legacy_expected_data_time_by_key: dict[AssetKey, Optional[datetime.datetime]] = {}
        self.legacy_data_time_resolver = CachingDataTimeResolver(self.instance_queryer)

        self.request_subsets_by_key: dict[EntityKey, EntitySubset] = {}

    @property
    def instance_queryer(self) -> "CachingInstanceQueryer":
        return self.asset_graph_view.get_inner_queryer_for_back_compat()

    @property
    def evaluation_time(self) -> datetime.datetime:
        return self.asset_graph_view.effective_dt

    @property
    def asset_graph(self) -> "BaseAssetGraph[BaseAssetNode]":
        return self.asset_graph_view.asset_graph

    @property
    def evaluated_asset_keys_and_parents(self) -> AbstractSet[AssetKey]:
        asset_keys = {ek for ek in self.entity_keys if isinstance(ek, AssetKey)}
        return {
            parent for ek in asset_keys for parent in self.asset_graph.get(ek).parent_keys
        } | asset_keys

    @property
    def asset_records_to_prefetch(self) -> Sequence[AssetKey]:
        return [key for key in self.evaluated_asset_keys_and_parents if self.asset_graph.has(key)]

    def prefetch(self) -> None:
        """Pre-populate the cached values here to avoid situations in which the new latest_storage_id
        value is calculated using information that comes in after the set of asset partitions with
        new parent materializations is calculated, as this can result in materializations being
        ignored if they happen between the two calculations.
        """
        self.logger.info(
            f"Prefetching asset records for {len(self.asset_records_to_prefetch)} records."
        )
        self.instance_queryer.prefetch_asset_records(self.asset_records_to_prefetch)
        self.logger.info("Done prefetching asset records.")

    def evaluate(self) -> tuple[Sequence[AutomationResult], Sequence[EntitySubset[EntityKey]]]:
        return asyncio.run(self.async_evaluate())

    async def async_evaluate(
        self,
    ) -> tuple[Sequence[AutomationResult], Sequence[EntitySubset[EntityKey]]]:
        self.prefetch()
        num_conditions = len(self.entity_keys)
        num_evaluated = 0

        async def _evaluate_entity_async(entity_key: EntityKey, offset: int):
            self.logger.debug(
                f"Evaluating {entity_key.to_user_string()} ({num_evaluated+offset}/{num_conditions})"
            )

            try:
                await self.evaluate_entity(entity_key)
            except Exception as e:
                raise Exception(
                    f"Error while evaluating conditions for {entity_key.to_user_string()}"
                ) from e

            result = self.current_results_by_key[entity_key]
            num_requested = result.true_subset.size
            if result.true_subset.is_partitioned:
                requested_str = ",".join(result.true_subset.expensively_compute_partition_keys())
            else:
                requested_str = "(no partition)"
            log_fn = self.logger.info if num_requested > 0 else self.logger.debug
            log_fn(
                f"{entity_key.to_user_string()} evaluation result: {num_requested} "
                f"requested ({requested_str}) "
                f"({format(result.end_timestamp - result.start_timestamp, '.3f')} seconds)"
            )

        for topo_level in self.asset_graph.toposorted_entity_keys_by_level:
            coroutines = [
                _evaluate_entity_async(entity_key, offset)
                for offset, entity_key in enumerate(topo_level)
                if entity_key in self.entity_keys
            ]
            await asyncio.gather(*coroutines)
            num_evaluated += len(coroutines)

        return list(self.current_results_by_key.values()), [
            v for v in self.request_subsets_by_key.values() if not v.is_empty
        ]

    async def evaluate_entity(self, key: EntityKey) -> None:
        # evaluate the condition of this asset
        result = await AutomationContext.create(key=key, evaluator=self).evaluate_async()

        # update dictionaries to keep track of this result
        self.current_results_by_key[key] = result
        self._add_request_subset(result.true_subset)

        if isinstance(key, AssetKey):
            self.legacy_expected_data_time_by_key[key] = result.compute_legacy_expected_data_time()
            # handle cases where an entity must be materialized with others
            self._handle_execution_set(result)

    def _add_request_subset(self, subset: EntitySubset) -> None:
        """Adds the provided subset to the dictionary tracking what we will request on this tick."""
        if subset.key not in self.request_subsets_by_key:
            self.request_subsets_by_key[subset.key] = subset
        else:
            self.request_subsets_by_key[subset.key] = self.request_subsets_by_key[
                subset.key
            ].compute_union(subset)

    def _handle_execution_set(self, result: AutomationResult[AssetKey]) -> None:
        # if we need to materialize any partitions of a non-subsettable multi-asset, we need to
        # materialize all of them
        asset_key = result.key
        execution_set_keys = self.asset_graph.get(asset_key).execution_set_entity_keys

        if len(execution_set_keys) > 1 and result.true_subset.size > 0:
            for neighbor_key in execution_set_keys:
                if isinstance(neighbor_key, AssetKey):
                    self.legacy_expected_data_time_by_key[neighbor_key] = (
                        self.legacy_expected_data_time_by_key[asset_key]
                    )

                # make sure that the true_subset of the neighbor is accurate -- when it was
                # evaluated it may have had a different requested subset. however, because
                # all these neighbors must be executed as a unit, we need to union together
                # the subset of all required neighbors
                neighbor_true_subset = result.true_subset.compute_mapped_subset(
                    neighbor_key, direction="up"
                )
                if neighbor_key in self.current_results_by_key:
                    self.current_results_by_key[
                        neighbor_key
                    ].set_internal_serializable_subset_override(
                        neighbor_true_subset.convert_to_serializable_subset()
                    )

                self._add_request_subset(neighbor_true_subset)



from collections.abc import Iterable, Mapping, Sequence
from typing import Any, Optional, TypeVar, cast

from dagster import _check as check
from dagster._config import Shape
from dagster._core.definitions.resource_requirement import ResourceAddable
from dagster._core.definitions.utils import DEFAULT_IO_MANAGER_KEY
from dagster._core.errors import DagsterInvalidConfigError, DagsterInvalidInvocationError
from dagster._core.execution.build_resources import wrap_resources_for_execution
from dagster._utils.merger import merge_dicts

T = TypeVar("T", bound=ResourceAddable)


def with_resources(
    definitions: Iterable[T],
    resource_defs: Mapping[str, object],
    resource_config_by_key: Optional[Mapping[str, Any]] = None,
) -> Sequence[T]:
    """Adds dagster resources to copies of resource-requiring dagster definitions.

    An error will be thrown if any provided definitions have a conflicting
    resource definition provided for a key provided to resource_defs. Resource
    config can be provided, with keys in the config dictionary corresponding to
    the keys for each resource definition. If any definition has unsatisfied
    resource keys after applying with_resources, an error will be thrown.

    Args:
        definitions (Iterable[ResourceAddable]): Dagster definitions to provide resources to.
        resource_defs (Mapping[str, object]):
            Mapping of resource keys to objects to satisfy
            resource requirements of provided dagster definitions.
        resource_config_by_key (Optional[Mapping[str, Any]]):
            Specifies config for provided resources. The key in this dictionary
            corresponds to configuring the same key in the resource_defs
            dictionary.

    Examples:
        .. code-block:: python

            from dagster import asset, resource, with_resources

            @resource(config_schema={"bar": str})
            def foo_resource():
                ...

            @asset(required_resource_keys={"foo"})
            def asset1(context):
                foo = context.resources.foo
                ...

            @asset(required_resource_keys={"foo"})
            def asset2(context):
                foo = context.resources.foo
                ...

            asset1_with_foo, asset2_with_foo = with_resources(
                [asset1, asset2],
                resource_defs={
                    "foo": foo_resource
                },
                resource_config_by_key={
                    "foo": {
                        "config": {"bar": ...}
                    }
                }
            )
    """
    from dagster._config import validate_config
    from dagster._core.definitions.job_definition import (
        default_job_io_manager_with_fs_io_manager_schema,
    )

    check.mapping_param(resource_defs, "resource_defs")
    resource_config_by_key = check.opt_mapping_param(
        resource_config_by_key, "resource_config_by_key"
    )

    resource_defs = wrap_resources_for_execution(
        merge_dicts(
            {DEFAULT_IO_MANAGER_KEY: default_job_io_manager_with_fs_io_manager_schema},
            resource_defs,
        )
    )

    for key, resource_def in resource_defs.items():
        if key in resource_config_by_key:
            resource_config = resource_config_by_key[key]
            if not isinstance(resource_config, dict) or "config" not in resource_config:
                raise DagsterInvalidInvocationError(
                    f"Error with config for resource key '{key}': Expected a "
                    "dictionary of the form {'config': ...}, but received "
                    f"{resource_config}"
                )

            outer_config_shape = Shape({"config": resource_def.get_config_field()})
            config_evr = validate_config(outer_config_shape, resource_config)
            if not config_evr.success:
                raise DagsterInvalidConfigError(
                    f"Error when applying config for resource with key '{key}' ",
                    config_evr.errors,
                    resource_config,
                )
            resource_defs[key] = resource_defs[key].configured(resource_config["config"])

    transformed_defs: list[T] = []
    for definition in definitions:
        transformed_defs.append(cast(T, definition.with_resources(resource_defs)))

    return transformed_defs
